import math
import os
import sys

import cv2
import json
import numpy as np

from PyQt5.QtCore import Qt, QSize, QTimer
from PyQt5.QtWidgets import QApplication, QCheckBox, QFileDialog, QMessageBox, QLabel, QSlider, QStyleOptionSlider, QVBoxLayout

import zebrazoom.code.paths as paths
import zebrazoom.code.util as util
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
from zebrazoom.code.createValidationVideo import calculateInfoFrame, drawInfoFrame
from zebrazoom.code.getHyperparameters import getHyperparametersSimple
from zebrazoom.code.preprocessImage import preprocessImage


def getFramesCallback(videoPath, folderName, numWell, numAnimal, zoom, start, framesToShow=0, ZZoutputLocation='', supstruct=None, config=None):
  s1  = "ZZoutput"
  s2  = folderName
  s3b = "results_"
  s4  = folderName
  s5  = ".avi"
  s5b = ".txt"

  if len(ZZoutputLocation):
    initialPath = ZZoutputLocation
  else:
    initialPath = paths.getDefaultZZoutputFolder()

  if config is None:
    with open(os.path.join(initialPath, os.path.join(s2, 'configUsed.json'))) as f:
      config = json.load(f)
  hyperparameters = getHyperparametersSimple(config)

  if os.path.splitext(folderName)[1] != '.h5':
    folderPath = os.path.join(initialPath, folderName)
    resultFile = next((f for f in os.listdir(folderPath) if os.path.isfile(os.path.join(folderPath, f)) and os.path.splitext(f)[1] == '.txt' and f.startswith('results_')), None)
    if resultFile is None:
      return
    resultsPath = os.path.join(initialPath, folderName, resultFile)
    if supstruct is None:
      with open(resultsPath) as f:
        supstruct = json.load(f)
  else:
    resultsPath = os.path.join(initialPath, folderName)
  
  if "pathToOriginalVideo" in supstruct:
    videoPath = supstruct["pathToOriginalVideo"]
    if not os.path.exists(videoPath):
      app = QApplication.instance()
      if QMessageBox.critical(app.window, "Video not found", "Cannot display the validation video because the video used to run the tracking can no longer be found. Would you like to update the video path stored in the results file?",
                              buttons=QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No, defaultButton=QMessageBox.StandardButton.Yes) != QMessageBox.StandardButton.Yes:
        return None
      videoName, _ = QFileDialog.getOpenFileName(app.window, 'Select video', os.path.expanduser("~"))
      if not videoName:
        return
      videoPath = supstruct["pathToOriginalVideo"] = videoName
      if os.path.splitext(resultsPath)[1] != '.h5':
        with open(resultsPath, 'w') as f:
          json.dump(supstruct, f)
      else:
        import h5py
        with h5py.File(resultsPath, 'a') as results:
          results.attrs['pathToOriginalVideo'] = videoName
  else:
    if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] and os.path.exists(os.path.join(initialPath, os.path.join(s1, os.path.join(s2, 'originalVideoWithoutAnyTrackingDisplayed_pleaseUseTheGUIToVisualizeTrackingPoints.avi')))):
      # The "exist" check above is only to insure compatibility with videos tracked prior to this update
      if os.path.splitext(resultsPath)[1] != '.h5':
        videoPath = os.path.join(initialPath, os.path.join(s2, 'originalVideoWithoutAnyTrackingDisplayed_pleaseUseTheGUIToVisualizeTrackingPoints.avi'))
      else:
        videoPath = f'{os.path.splitext(resultsPath)[0]}_originalVideoWithoutAnyTrackingDisplayed_pleaseUseTheGUIToVisualizeTrackingPoints.avi'
    else:
      if os.path.splitext(resultsPath)[1] != '.h5':
        videoPath = os.path.join(initialPath, os.path.join(s2, s4 + s5))
      else:
        videoPath = f'{os.path.splitext(resultsPath)[0]}.avi'

  if not(os.path.exists(videoPath)) and os.path.splitext(resultsPath)[1] != '.h5':
    mypath = os.path.join(initialPath, s2)
    onlyfiles = [f for f in os.listdir(mypath) if os.path.isfile(os.path.join(mypath, f))]
    resultFile = ''
    for fileName in onlyfiles:
      if '.avi' in fileName:
        resultFile = fileName
    videoPath = os.path.join(initialPath, os.path.join(s2, resultFile))

  if not os.path.isfile(videoPath):
    app = QApplication.instance()
    QMessageBox.critical(app.window, "Video not found", "Cannot display the validation video because it could not be found.")
    return

  cap = zzVideoReading.VideoCapture(videoPath)

  nx    = int(cap.get(3))
  ny    = int(cap.get(4))
  max_l = int(cap.get(7))
  if max_l == 1:
    return None
  frameRange = (supstruct["firstFrame"], supstruct["lastFrame"] - 1) if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] or "pathToOriginalVideo" in supstruct else (0, max_l -1)

  if not("firstFrame" in supstruct):
    supstruct["firstFrame"] = 1
    print("supstruct['firstFrame'] not found")

  infoWells = []

  HeadX = np.zeros(max_l + supstruct["firstFrame"])
  HeadY = np.zeros(max_l + supstruct["firstFrame"])

  if ((numWell != -1) and (zoom)):
    lastEnd = 0
    lastXpos = supstruct["wellPoissMouv"][numWell][numAnimal][0]["HeadX"][0]
    lastYpos = supstruct["wellPoissMouv"][numWell][numAnimal][0]["HeadY"][0]
    for k in range(0,len(supstruct["wellPoissMouv"][numWell][numAnimal])):
      beg = supstruct["wellPoissMouv"][numWell][numAnimal][k]["BoutStart"]
      end = supstruct["wellPoissMouv"][numWell][numAnimal][k]["BoutEnd"]
      for l in range(lastEnd, beg):
        HeadX[l] = lastXpos
        HeadY[l] = lastYpos
      for l in range(beg, end):
        HeadX[l]  = supstruct["wellPoissMouv"][numWell][numAnimal][k]["HeadX"][l-beg]
        HeadY[l]  = supstruct["wellPoissMouv"][numWell][numAnimal][k]["HeadY"][l-beg]
      lastEnd = end
      lastXpos = supstruct["wellPoissMouv"][numWell][numAnimal][k]["HeadX"][end-1-beg]
      lastYpos = supstruct["wellPoissMouv"][numWell][numAnimal][k]["HeadY"][end-1-beg]

    for l in range(lastEnd, max_l + supstruct["firstFrame"]):
      HeadX[l] = lastXpos
      HeadY[l] = lastYpos

  # /* Getting the info about well positions */
  analyzeAllWellsAtTheSameTime = 0
  if (analyzeAllWellsAtTheSameTime == 0):
    for i in range(0, len(supstruct["wellPositions"])):
      x = 0
      y = 0
      lengthX = 0
      lengthY = 0
      rectangleWellArea = 1
      if (rectangleWellArea == 0): # circular wells
        x = supstruct["wellPositions"][i]["topLeftX"]
        y = supstruct["wellPositions"][i]["topLeftY"]
        r = supstruct["wellPositions"][i]["diameter"]
        lengthX = 300 # wellOutputVideoDiameter;
        lengthY = 300 # wellOutputVideoDiameter;
      else:
        x = supstruct["wellPositions"][i]["topLeftX"]
        y = supstruct["wellPositions"][i]["topLeftY"]
        lengthX = supstruct["wellPositions"][i]["lengthX"]
        lengthY = supstruct["wellPositions"][i]["lengthY"]
      if (x < 0):
        x = 0
      if (y < 0):
        y = 0
      infoWells.append([x, y, lengthX, lengthY])
  else:
    infoWells.append([0, 0, nx, ny])
  
  infoFrame = None
  if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] or "pathToOriginalVideo" in supstruct:
    infoFrame, colorModifTab = calculateInfoFrame(supstruct, hyperparameters, max_l)
  
  x = 0
  y = 0
  lengthX = 0
  lengthY = 0
  if (numWell != -1):
    x = infoWells[numWell][0]
    y = infoWells[numWell][1]
    lengthX = infoWells[numWell][2]
    lengthY = infoWells[numWell][3]
  else:
    lengthX = nx
    lengthY = ny

  l = start - supstruct["firstFrame"] + 1 if start > 0 else 0
  if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] or "pathToOriginalVideo" in supstruct:
    l += supstruct["firstFrame"]

  xOriginal = x
  yOriginal = y

  def getFrame(frameSlider, timer=None, plotTrackingPointsCheckbox=None, stopTimer=True):
    nonlocal x
    nonlocal y
    nonlocal lengthX
    nonlocal lengthY

    l = frameSlider.value()
    if timer is not None and timer.isActive() and (l == frameSlider.maximum() or stopTimer):
      timer.stop()

    cap.set(1, l)
    ret, img = cap.read()
    
    if hyperparameters["imagePreProcessMethod"]:
      img = preprocessImage(img, hyperparameters)
    
    if infoFrame is not None and plotTrackingPointsCheckbox is not None and plotTrackingPointsCheckbox.isChecked():
      drawInfoFrame(l, img, infoFrame, colorModifTab, hyperparameters)

    if numWell != -1 and zoom:
      length = 250
      xmin = int(HeadX[l + supstruct["firstFrame"] - 1] - length/2)
      xmax = int(HeadX[l + supstruct["firstFrame"] - 1] + length/2)
      ymin = int(HeadY[l + supstruct["firstFrame"] - 1] - length/2)
      ymax = int(HeadY[l + supstruct["firstFrame"] - 1] + length/2)

      x = max(xmin + xOriginal, 0)
      y = max(ymin + yOriginal, 0)
      lengthX = xmax - xmin
      lengthY = ymax - ymin

      if y + lengthY >= len(img):
        lengthY = len(img) - y - 1
      if x + lengthX >= len(img[0]):
        lengthX = len(img[0]) - x - 1

    if (numWell != -1):
      img = img[y:y+lengthY, x:x+lengthX]

    frameNumber = l if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] or "pathToOriginalVideo" in supstruct else (l + supstruct["firstFrame"])
    if lengthX > 100 and lengthY > 100:
      font = cv2.FONT_HERSHEY_SIMPLEX
      cv2.putText(img,str(frameNumber),(int(lengthX-110), int(lengthY-30)),font,1,(0,255,0))
    else:
      blank_image = np.zeros((len(img)+30, len(img[0]), 3), np.uint8)
      blank_image[0:len(img), 0:len(img[0])] = img
      img = blank_image
      font = cv2.FONT_HERSHEY_SIMPLEX
      cv2.putText(img, str(frameNumber), (int(0), int(lengthY+25)), font, 1, (0,255,0))

    return img

  wellShape = None if config.get("noWellDetection", False) or (hyperparameters["headEmbeded"] and not hyperparameters["oneWellManuallyChosenTopLeft"]) else 'rectangle' if config.get("wellsAreRectangles", False) or len(config.get("oneWellManuallyChosenTopLeft", '')) or int(config.get("multipleROIsDefinedDuringExecution", 0)) or config.get("groupOfMultipleSameSizeAndShapeEquallySpacedWells", False) else 'circle'
  return getFrame, frameRange, l, infoFrame is not None, supstruct['wellPositions'], wellShape


def readValidationVideo(videoPath, folderName, numWell, numAnimal, zoom, start, framesToShow=0, ZZoutputLocation='', supstruct=None, config=None):
  frameInfo = getFramesCallback(videoPath, folderName, numWell, numAnimal, zoom, start, framesToShow=framesToShow, ZZoutputLocation=ZZoutputLocation, supstruct=supstruct, config=config)
  if frameInfo is None:
    return
  getFrame, frameRange, frame, toggleTrackingPoints, _, _ = frameInfo
  layout = QVBoxLayout()

  video = QLabel()
  layout.addWidget(video, stretch=1)

  frameSlider = QSlider(Qt.Orientation.Horizontal)
  frameSlider.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
  frameSlider.setPageStep(50)
  frameSlider.setRange(*frameRange)
  frameSlider.setValue(frame)
  frameSlider.valueChanged.connect(lambda: util.setPixmapFromCv(getFrame(frameSlider, timer, plotTrackingPointsCheckbox, stopTimer), video))
  layout.addWidget(frameSlider)
  shortcutsLabel = QLabel("Left Arrow, Right Arrow, Page Up, Page Down, Home and End keys can be used to navigate through the video.")
  shortcutsLabel.setWordWrap(True)
  layout.addWidget(shortcutsLabel)
  plotTrackingPointsCheckbox = None
  if toggleTrackingPoints:
    plotTrackingPointsCheckbox = QCheckBox("Display tracking points")
    plotTrackingPointsCheckbox.setChecked(True)
    plotTrackingPointsCheckbox.toggled.connect(lambda: timer.isActive() or util.setPixmapFromCv(getFrame(frameSlider, timer, plotTrackingPointsCheckbox, stopTimer), video))
    layout.addWidget(plotTrackingPointsCheckbox)

  stopTimer = True
  timer = QTimer()
  timer.setInterval(1)

  def nextFrame():
    nonlocal stopTimer
    stopTimer = False
    frameSlider.setValue(frameSlider.value() + 1)
    stopTimer = True
  timer.timeout.connect(nextFrame)

  startFrame = getFrame(frameSlider, timer, plotTrackingPointsCheckbox, stopTimer)
  timer.start()
  util.showDialog(layout, title="Video", labelInfo=(startFrame, video))
  timer.stop()
  del getFrame



import struct
import csv
import mmap
import datetime
import array
import logging
import os
import csv
import glob
import time
import configparser
import numpy as np
import cv2
from pathlib import Path
import platform

class ZzVideoReading():

  def __init__(self, videoPath):
  
    seq_path  = videoPath

    cfg = configparser.ConfigParser()
    cfg.read(seq_path)
    
    width = int(cfg.get('Sequence Settings', 'Width'))
    height = int(cfg.get('Sequence Settings', 'Height'))
    bpp = int(cfg.get('Sequence Settings', 'BytesPerPixel'))
    num_images = cfg.get('Sequence Settings', 'Number of files')
    bin_file = cfg.get('Sequence Settings', 'Bin File')
    sqb_path = seq_path.replace('.seq', '.sqb')
    
    pathstr = os.path.dirname(seq_path)
    
    f = open(sqb_path, 'rb')
    
    self.sqb_path = sqb_path
    self.width = width
    self.height = height
    self.bpp = bpp
    self.num_images = int(num_images)
    self.bin_file = bin_file
    self.sqb_path = sqb_path
    self.pathstr = pathstr
    self.f = f
    self.lastFrameRead = -1
  
  def get(self, idOfInfoRequested):
    
    if idOfInfoRequested == 1:
      return self.lastFrameRead + 1
    elif idOfInfoRequested == 3:
      return self.width
    elif idOfInfoRequested == 4:
      return self.height
    elif idOfInfoRequested == 7:
      return self.num_images
    elif idOfInfoRequested == 5:
      return 24 # fake input video fps
  
  def isOpened(self):
    return True
  
  def release(self):
    del self
    
  def read(self):
    
    if self.lastFrameRead + 1 < self.num_images:
      
      try:
        offset = struct.unpack('i', self.f.read(4))
        padding = self.f.read(4)
        timestamp = struct.unpack('d', self.f.read(8))
        binfile = struct.unpack('i', self.f.read(4))
        padding = self.f.read(4)
      except:
        print("Problem with Hiris video format", self.lastFrameRead + 1)
        return [False, []]
      
      bin_path = os.path.join("%s" % (self.pathstr), "%s%0.5d.bin" % (self.bin_file, binfile[0]))
      
      if not(os.path.exists(bin_path)):
        print("Hiris video format: frame not found:", self.lastFrameRead + 1)
        return [False, []]
      
      f_bin = open(bin_path, 'rb')
      f_bin.seek(offset[0], os.SEEK_SET)
      
      bytes = f_bin.read(self.height*self.width*self.bpp)
      
      if self.bpp == 2:
        buffer = np.frombuffer(bytes, dtype=np.uint16)
      else:
        buffer = np.frombuffer(bytes, dtype=np.uint8)
      
      
      if len(buffer) == 3*self.height*self.width:
        nparr2 = buffer.reshape(self.height, self.width, 3)
      else:
        nparr2 = buffer.reshape(self.height, self.width)
        nparr2 = cv2.cvtColor(nparr2, cv2.COLOR_GRAY2RGB)
      
      self.lastFrameRead = self.lastFrameRead + 1
      
      return [True, nparr2]
    
    else:
      
      return [False, []]
  
  
  def set(self, propToChange, numImage):
    
    if propToChange == 1:
      
      if numImage <= 0:
        
        self.f.close()
        self.f = open(self.sqb_path, 'rb')
        self.lastFrameRead = -1
      
      elif numImage >= self.num_images:
        
        self.lastFrameRead = int(self.num_images)
        
      elif numImage > self.lastFrameRead:
        
        justToMove = self.f.read(24 * (int(numImage) - int(self.lastFrameRead) - 1))
        self.lastFrameRead = numImage - 1  
        
      elif numImage <= self.lastFrameRead:
        
        self.f.close()
        self.f = open(self.sqb_path, 'rb')
        
        justToMove = self.f.read(24 * int(numImage))
        self.lastFrameRead = numImage - 1


def VideoCapture(videoPath):
  
  if '.seq' in videoPath:
    
    zzVidCapture = ZzVideoReading(videoPath)
    
    return zzVidCapture
  
  elif '.sqb' in videoPath:
    
    zzVidCapture = ZzVideoReading(videoPath.replace('.sqb', '.seq'))
    
    return zzVidCapture
  
  else:
    
    return cv2.VideoCapture(videoPath)



from zebrazoom.code.dataPostProcessing.createPandasDataFrameOfParameters import createPandasDataFrameOfParameters
import h5py
import pandas as pd
import json
import os


def _getParametersUsedForCalculation(path, trialID, ZZoutputLocation):
  rootPath = os.path.join(ZZoutputLocation if path == "defaultZZoutputFolder" else path, trialID)
  if os.path.splitext(trialID)[1] != '.h5':
    parametersPath = os.path.join(rootPath, f'{trialID}.pkl')
    if not os.path.exists(parametersPath):
      return {}
    parametersUsedForCalculationPath = os.path.join(rootPath, 'parametersUsedForCalculation.json')
    if not os.path.exists(parametersUsedForCalculationPath):
      return {}
    with open(parametersUsedForCalculationPath) as f:
      return json.load(f)
  else:
    with h5py.File(rootPath, 'r') as results:
      videoFPS = results.attrs.get('videoFPS', None)
      videoPixelSize = results.attrs.get('videoPixelSize', None)
      frameStepForDistanceCalculation = results.attrs.get('frameStepForDistanceCalculation', None)
    parametersUsedForCalculation = {}
    if videoFPS is not None:
      parametersUsedForCalculation['videoFPS'] = float(videoFPS)
    if videoPixelSize is not None:
      parametersUsedForCalculation['videoPixelSize'] = float(videoPixelSize)
    if frameStepForDistanceCalculation is not None:
      parametersUsedForCalculation['frameStepForDistanceCalculation'] = int(frameStepForDistanceCalculation)
    return parametersUsedForCalculation


def generatePklDataFileForVideo(excelFileName, ZZoutputLocation, frameStepForDistanceCalculation, forcePandasRecreation=0, reusingParametersCb=None):
  
  # Generate .pkl data file inside the result video folder if it doesn't already exist
  excelFile = pd.read_excel(excelFileName)
  
  for videoId in range(0, len(excelFile)):
    # If it exists, retrives the frameStepForDistanceCalculationUsed parameter previously used to calculate the distance travelled.
    newParametersUsedForCalculation = {'videoFPS': float(excelFile.loc[videoId, 'fq']), 'videoPixelSize': float(excelFile.loc[videoId, 'pixelsize']), 'frameStepForDistanceCalculation': int(frameStepForDistanceCalculation)}
    parametersUsedForCalculation = _getParametersUsedForCalculation(excelFile.loc[videoId, 'path'], excelFile.loc[videoId, 'trial_id'], ZZoutputLocation)
    
    # Generates the .pkl data file if it doesn't already exist OR if a frameStepForDistanceCalculationUsed different than before is requested or if a videoFPS or videoPixelSize different than before is used
    regenerate = forcePandasRecreation or parametersUsedForCalculation != newParametersUsedForCalculation
    if not regenerate and reusingParametersCb is not None:
      forcePandasRecreation = reusingParametersCb()
      regenerate = forcePandasRecreation
      reusingParametersCb = None

    if regenerate and os.path.splitext(excelFile.loc[videoId, 'trial_id'])[1] != '.h5':
      print("Generating pkl datafile for video " + excelFile.loc[videoId, 'trial_id'] + " , because:")
      if not(os.path.exists(os.path.join(os.path.join(ZZoutputLocation if excelFile.loc[videoId, 'path'] == "defaultZZoutputFolder" else excelFile.loc[videoId, 'path'], excelFile.loc[videoId, 'trial_id']), excelFile.loc[videoId, 'trial_id'] + '.pkl'))):
        print("the path", os.path.join(os.path.join(ZZoutputLocation if excelFile.loc[videoId, 'path'] == "defaultZZoutputFolder" else excelFile.loc[videoId, 'path'], excelFile.loc[videoId, 'trial_id']), excelFile.loc[videoId, 'trial_id'] + '.pkl'), "does not exist")
      if not(len(parametersUsedForCalculation)):
        print("len(parametersUsedForCalculation) is equal to", len(parametersUsedForCalculation))
      else:
        if int(frameStepForDistanceCalculation) != int(parametersUsedForCalculation['frameStepForDistanceCalculation']):
          print("int(frameStepForDistanceCalculation)=", int(frameStepForDistanceCalculation), "and: int(parametersUsedForCalculation['frameStepForDistanceCalculation']=", int(parametersUsedForCalculation['frameStepForDistanceCalculation']))
        if float(excelFile.loc[videoId, 'fq']) != float(parametersUsedForCalculation['videoFPS']):
          print("float(excelFile.loc[videoId, 'fq'])", float(excelFile.loc[videoId, 'fq']), "float(parametersUsedForCalculation['videoFPS']):", float(parametersUsedForCalculation['videoFPS']))
        if float(excelFile.loc[videoId, 'pixelsize']) != float(parametersUsedForCalculation['videoFPS']):
          print("float(excelFile.loc[videoId, 'pixelsize'])", float(excelFile.loc[videoId, 'pixelsize']), "float(parametersUsedForCalculation['videoPixelSize'])", float(parametersUsedForCalculation['videoPixelSize']))
      
      videoName      = excelFile.loc[videoId, 'trial_id']
      videoExtension = os.path.splitext(os.path.split(excelFileName)[1])[1]
      hyperparameters = {}
      hyperparameters["nbWells"] = len(excelFile.loc[videoId, 'condition'][1:-1].split(','))
      hyperparameters["frameStepForDistanceCalculation"] = frameStepForDistanceCalculation
      hyperparameters["videoFPS"]       = float(excelFile.loc[videoId, 'fq'])
      hyperparameters["videoPixelSize"] = float(excelFile.loc[videoId, 'pixelsize'])
      createPandasDataFrameOfParameters(hyperparameters, videoName, videoExtension, ZZoutputLocation if excelFile.loc[videoId, 'path'] == "defaultZZoutputFolder" else excelFile.loc[videoId, 'path'])
  return forcePandasRecreation


from zebrazoom.code.dataPostProcessing.createPandasDataFrameOfParameters import createPandasDataFrameOfParameters
from zebrazoom.code.getHyperparameters import getHyperparametersSimple

from ._createSuperStructFromH5 import createSuperStructFromH5
from ._openResultsFile import openResultsFile


def getKinematicParametersPerBout(videoName: str, numWell: int, numAnimal: int, numBout: int) -> dict:
  with openResultsFile(videoName, 'r+') as results:
    animalPath = f'dataForWell{numWell}/dataForAnimal{numAnimal}'
    if animalPath not in results:
      raise ValueError(f"data for animal {numAnimal} in well {numWell} doesn't exist")
    animalGroup = results[animalPath]
    if 'kinematicParametersPerBout' not in animalGroup:
      if 'videoFPS' not in results.attrs:
        raise ValueError(f'videoFPS not found in the results, cannot calculate kinematic parameters')
      if 'videoPixelSize' not in results.attrs:
        raise ValueError(f'videoPixelSize not found in the results, cannot calculate kinematic parameters')
      print(f'calculating and storing kinematic parameters for well {numWell}, animal {numAnimal}, bout {numBout}')
      superStruct = createSuperStructFromH5(results)
      hyperparameters = getHyperparametersSimple(dict(results['configurationFileUsed'].attrs))
      hyperparameters['H5filename'] = results.filename
      hyperparameters['videoFPS'] = results.attrs['videoFPS']
      hyperparameters['videoPixelSize'] = results.attrs['videoPixelSize']
      createPandasDataFrameOfParameters(hyperparameters, '', '', '', superStruct)

    dataset = animalGroup['kinematicParametersPerBout']
    numberOfBouts, = dataset.shape
    if numBout >= numberOfBouts:
      raise ValueError(f"cannot get data for bout {numBout}, total number of detected bouts is {numberOfBouts}")
    return {col: dataset[col][numBout] for col in dataset.attrs['columns']}



import numpy as np
from scipy import ndimage

def applyMedianFilterOnCurvature(curvatureValues: np.array, rolling_window: int=3) -> np.array:
  
  curvatureValues = ndimage.median_filter(curvatureValues, size=rolling_window)
  
  return curvatureValues



import matplotlib.pyplot as plt
import math

def generateAllTimeTailAngleGraph(path, superStruct, generateAllTimeTailAngleGraphLineWidth):
  if not os.path.exists(path):
    os.makedirs(path)
  for i in range(0, len(superStruct["wellPoissMouv"])):
    for j in range(0, len(superStruct["wellPoissMouv"][i])):
      for k in range(0, len(superStruct["wellPoissMouv"][i][j])):
        tailAngle = superStruct["wellPoissMouv"][i][j][k]["TailAngle_smoothed"]
        plt.figure()
        tailAngle2 = [(180/math.pi)*t for t in tailAngle]
        plt.plot(tailAngle2, linewidth=generateAllTimeTailAngleGraphLineWidth)
        plt.axis([0, 1300, -38, 38])
        plt.savefig(path+'/well'+str(i)+'_bout'+str(k)+'.png', dpi=1200)
        plt.savefig(path+'/well'+str(i)+'_bout'+str(k)+'.pdf', dpi=1200)
        plt.savefig(path+'/well'+str(i)+'_bout'+str(k)+'.eps', dpi=1200)



import cv2
import contextlib
import re
import os
import json
import shutil
import sys
import subprocess
from functools import partial
from multiprocessing import Pool

from matplotlib.figure import Figure
import math
import scipy.io as sio
import pandas as pd
from zebrazoom.code.vars import getGlobalVariables
globalVariables = getGlobalVariables()

from zebrazoom.zebraZoomVideoAnalysis import ZebraZoomVideoAnalysis
from zebrazoom.getTailExtremityFirstFrame import getTailExtremityFirstFrame
import zebrazoom.code.paths as paths
import zebrazoom.code.util as util

from PyQt5.QtCore import Qt, QAbstractTableModel, QItemSelection, QItemSelectionModel, QModelIndex, QSize, QUrl
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtWidgets import QAbstractItemView, QApplication, QCheckBox, QFileDialog, QFileSystemModel, QFrame, QHBoxLayout, QHeaderView, QLabel, QLineEdit, QListView, QMessageBox, QPushButton, QScrollArea, QSpacerItem, QSpinBox, QTableView, QTextEdit, QTreeView, QVBoxLayout, QWidget


LARGE_FONT= ("Verdana", 12)
_VIDEO_EXTENSIONS = {'.264', '.3g2', '.3gp', '.3gp2', '.3gpp', '.3gpp2', '.3mm', '.3p2', '.60d', '.787', '.89', '.aaf', '.aec', '.aep', '.aepx', '.aet', '.aetx', '.ajp',
                     '.ale', '.am', '.amc', '.amv', '.amx', '.anim', '.aqt', '.arcut', '.arf', '.asf', '.asx', '.avb', '.avc', '.avd', '.avi', '.avp', '.avs', '.avs',
                     '.avv', '.axm', '.bdm', '.bdmv', '.bdt2', '.bdt3', '.bik', '.bix', '.bmk', '.bnp', '.box', '.bs4', '.bsf', '.bvr', '.byu', '.camproj', '.camrec',
                     '.camv', '.ced', '.cel', '.cine', '.cip', '.clpi', '.cmmp', '.cmmtpl', '.cmproj', '.cmrec', '.cpi', '.cst', '.cvc', '.cx3', '.d2v', '.d3v', '.dat',
                     '.dav', '.dce', '.dck', '.dcr', '.dcr', '.ddat', '.dif', '.dir', '.divx', '.dlx', '.dmb', '.dmsd', '.dmsd3d', '.dmsm', '.dmsm3d', '.dmss', '.dmx', '.dnc',
                     '.dpa', '.dpg', '.dream', '.dsy', '.dv', '.dv-avi', '.dv4', '.dvdmedia', '.dvr', '.dvr-ms', '.dvx', '.dxr', '.dzm', '.dzp', '.dzt', '.edl', '.evo', '.eye',
                     '.ezt', '.f4p', '.f4v', '.fbr', '.fbr', '.fbz', '.fcp', '.fcproject', '.ffd', '.flc', '.flh', '.fli', '.flv', '.flx', '.gfp', '.gl', '.gom', '.grasp',
                     '.gts', '.gvi', '.gvp', '.h264', '.hdmov', '.hkm', '.ifo', '.imovieproj', '.imovieproject', '.ircp', '.irf', '.ism', '.ismc', '.ismv', '.iva', '.ivf',
                     '.ivr', '.ivs', '.izz', '.izzy', '.jss', '.jts', '.jtv', '.k3g', '.kmv', '.ktn', '.lrec', '.lsf', '.lsx', '.m15', '.m1pg', '.m1v', '.m21', '.m21', '.m2a',
                     '.m2p', '.m2t', '.m2ts', '.m2v', '.m4e', '.m4u', '.m4v', '.m75', '.mani', '.meta', '.mgv', '.mj2', '.mjp', '.mjpg', '.mk3d', '.mkv', '.mmv', '.mnv',
                     '.mob', '.mod', '.modd', '.moff', '.moi', '.moov', '.mov', '.movie', '.mp21', '.mp21', '.mp2v', '.mp4', '.mp4v', '.mpe', '.mpeg', '.mpeg1', '.mpeg4',
                     '.mpf', '.mpg', '.mpg2', '.mpgindex', '.mpl', '.mpl', '.mpls', '.mpsub', '.mpv', '.mpv2', '.mqv', '.msdvd', '.mse', '.msh', '.mswmm', '.mts', '.mtv',
                     '.mvb', '.mvc', '.mvd', '.mve', '.mvex', '.mvp', '.mvp', '.mvy', '.mxf', '.mxv', '.mys', '.ncor', '.nsv', '.nut', '.nuv', '.nvc', '.ogm', '.ogv', '.ogx',
                     '.osp', '.otrkey', '.pac', '.par', '.pds', '.pgi', '.photoshow', '.piv', '.pjs', '.playlist', '.plproj', '.pmf', '.pmv', '.pns', '.ppj', '.prel', '.pro',
                     '.prproj', '.prtl', '.psb', '.psh', '.pssd', '.pva', '.pvr', '.pxv', '.qt', '.qtch', '.qtindex', '.qtl', '.qtm', '.qtz', '.r3d', '.rcd', '.rcproject',
                     '.rdb', '.rec', '.rm', '.rmd', '.rmd', '.rmp', '.rms', '.rmv', '.rmvb', '.roq', '.rp', '.rsx', '.rts', '.rts', '.rum', '.rv', '.rvid', '.rvl', '.sbk',
                     '.sbt', '.scc', '.scm', '.scm', '.scn', '.screenflow', '.sec', '.sedprj', '.seq', '.sfd', '.sfvidcap', '.siv', '.smi', '.smi', '.smil', '.smk', '.sml',
                     '.smv', '.spl', '.sqz', '.srt', '.ssf', '.ssm', '.stl', '.str', '.stx', '.svi', '.swf', '.swi', '.swt', '.tda3mt', '.tdx', '.thp', '.tivo', '.tix',
                     '.tod', '.tp', '.tp0', '.tpd', '.tpr', '.trp', '.ts', '.tsp', '.ttxt', '.tvs', '.usf', '.usm', '.vc1', '.vcpf', '.vcr', '.vcv', '.vdo', '.vdr', '.vdx',
                     '.veg','.vem', '.vep', '.vf', '.vft', '.vfw', '.vfz', '.vgz', '.vid', '.video', '.viewlet', '.viv', '.vivo', '.vlab', '.vob', '.vp3', '.vp6', '.vp7',
                     '.vpj', '.vro', '.vs4', '.vse', '.vsp', '.w32', '.wcp', '.webm', '.wlmp', '.wm', '.wmd', '.wmmp', '.wmv', '.wmx', '.wot', '.wp3', '.wpl', '.wtv', '.wve',
                     '.wvx', '.xej', '.xel', '.xesc', '.xfl', '.xlmv', '.xmv', '.xvid', '.y4m', '.yog', '.yuv', '.zeg', '.zm1', '.zm2', '.zm3', '.zmv'}


def chooseVideoToAnalyze(self, createValidationVideo, chooseFrames):
    videoName, _ = QFileDialog.getOpenFileName(self.window, 'Select video', os.path.expanduser("~"), filter=f'Videos ({" ".join("*%s" % ext for ext in _VIDEO_EXTENSIONS)});; All files (*.*)')
    if not videoName:
      return
    ZZargs = ([videoName],)
    ZZkwargs = {'createValidationVideo': createValidationVideo}

    if chooseFrames:
      def beginningAndEndChosen():
        if "firstFrame" in self.configFile:
          ZZkwargs['firstFrame'] = self.configFile["firstFrame"]
        if "lastFrame" in self.configFile:
          ZZkwargs['lastFrame'] = self.configFile["lastFrame"]
        self.configFile.clear()
        ZZkwargs['backgroundExtractionForceUseAllVideoFrames'] = backgroundExtractionForceUseAllVideoFramesCheckbox.isChecked()
        self.show_frame("ConfigFilePromp")
        self.window.centralWidget().layout().currentWidget().setArgs(ZZargs, ZZkwargs)
      layout = QVBoxLayout()
      backgroundExtractionForceUseAllVideoFramesCheckbox = QCheckBox("Use all frames to calculate background")
      backgroundExtractionForceUseAllVideoFramesCheckbox.setChecked(True)
      layout.addWidget(backgroundExtractionForceUseAllVideoFramesCheckbox, alignment=Qt.AlignmentFlag.AlignCenter)
      util.chooseBeginningPage(self, videoName, "Choose where the analysis of your video should start.", "Ok, I want the tracking to start at this frame!",
                               lambda: util.chooseEndPage(self, videoName, "Choose where the analysis of your video should end.", "Ok, I want the tracking to end at this frame!", beginningAndEndChosen),
                               additionalLayout=layout)
    else:
      self.show_frame("ConfigFilePromp")
      self.window.centralWidget().layout().currentWidget().setArgs(ZZargs, ZZkwargs)


class _VideoFilesModel(QFileSystemModel):
  def __init__(self):
    super().__init__()
    self.setReadOnly(False)
    self.setNameFilterDisables(False)
    self.setNameFilters(("*.csv",))

  def data(self, index, role=Qt.ItemDataRole.DisplayRole):
    data = super().data(index, role=role)
    if role == Qt.ItemDataRole.EditRole:
      return os.path.splitext(data)[0]
    return data

  def setData(self, index, value, role=None):
    extension = os.path.splitext(self.data(index))[1]
    return super().setData(index, "%s%s" % (value, extension), role=role)


class _TrackingConfigurationsTreeView(QTreeView):
  def __init__(self):
    super().__init__()
    self.sizeHint = lambda: QSize(150, 1)
    self.header().setSectionResizeMode(QHeaderView.ResizeMode.Fixed)
    self.setRootIsDecorated(False)

  def resizeEvent(self, evt):
    super().resizeEvent(evt)
    self.setColumnWidth(0, evt.size().width())

  def flags(self, index):
    return super().flags(index) | Qt.ItemFlag.ItemIsEditable


class _VideosModel(QAbstractTableModel):
  _COLUMN_TITLES = ["Video", "Config"]
  _DEFAULT_ZZOUTPUT = paths.getDefaultZZoutputFolder()

  def __init__(self, filename):
    super().__init__()
    data = pd.read_csv(filename).fillna('')
    self._videos = data['Video'].tolist()
    self._configs = data['Config'].tolist()

  def rowCount(self, parent=None):
    return len(self._videos)

  def columnCount(self, parent=None):
    return len(self._COLUMN_TITLES)

  def data(self, index, role=Qt.ItemDataRole.DisplayRole):
    if index.isValid() and role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole, Qt.ItemDataRole.ToolTipRole):
      return self._videos[index.row()] if index.column() == 0 else self._configs[index.row()]
    return None

  def headerData(self, col, orientation, role):
    if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
      return self._COLUMN_TITLES[col]
    return None

  def addVideos(self, videos):
    if videos is None:
      return
    videos = [video for video in videos if video not in self._videos]
    self.beginInsertRows(QModelIndex(), self.rowCount(), self.rowCount() + len(videos) - 1)
    self._videos.extend(videos)
    self._configs.extend([''] * len(videos))
    self.endInsertRows()

  def setConfigs(self, idxs, config):
    if config is None:
      return
    self.beginResetModel()
    for idx in idxs:
      self._configs[idx] = config
    self.endResetModel()

  def removeSelectedRows(self, idxs):
    self.beginResetModel()
    for idx in reversed(sorted(idxs)):
      del self._videos[idx]
      del self._configs[idx]
    self.endResetModel()

  def getData(self):
    return self._videos, self._configs

  def hasUnsavedChanges(self, filename):
    if not os.path.exists(filename):  # file was deleted
      return False
    fileData = pd.read_csv(filename).fillna('')
    return not (fileData['Video'].tolist() == self._videos and fileData['Config'].tolist() == self._configs)

  def saveFile(self, filename):
    pd.DataFrame(columns=_VideosModel._COLUMN_TITLES, data=zip(self._videos, self._configs)).to_csv(filename, index=False)


class _TrackingConfigurationsSelectionModel(QItemSelectionModel):
  def __init__(self, window, table, *args):
    super().__init__(*args)
    self._window = window
    self._table = table
    self._blockSelection = False

  def setCurrentIndex(self, index, command):
    if index != self.currentIndex() and self._table.model() is not None and self._table.model().hasUnsavedChanges(self.model().filePath(self.currentIndex())) and \
        QMessageBox.question(self._window, "Unsaved changes", "Are you sure you want to proceed? Unsaved changes will be lost.",
                             defaultButton=QMessageBox.StandardButton.No) != QMessageBox.StandardButton.Yes:
      self._blockSelection = True
      return None
    return super().setCurrentIndex(index, command)

  def select(self, index, command):
    if self._blockSelection:
      self._blockSelection = False
      return None
    return super().select(index, command)


class _VideoSelectionPage(QWidget):
  def __init__(self, ZZkwargs):
    super().__init__()
    self._ZZkwargs = ZZkwargs

    app = QApplication.instance()

    layout = QVBoxLayout()
    layout.addWidget(util.apply_style(QLabel("Select videos and corresponding config files"), font=app.title_font), alignment=Qt.AlignmentFlag.AlignCenter)

    if ZZkwargs.get('sbatchMode', False):
      replaceLayout = QHBoxLayout()
      replaceLayout.addWidget(QLabel('Replace'), alignment=Qt.AlignmentFlag.AlignCenter)
      self._originalLineEdit = QLineEdit()
      self._originalLineEdit.setText('//l2export/iss02.')
      replaceLayout.addWidget(self._originalLineEdit, alignment=Qt.AlignmentFlag.AlignCenter)
      replaceLayout.addWidget(QLabel('with'), alignment=Qt.AlignmentFlag.AlignCenter)
      self._replaceLineEdit = QLineEdit()
      self._replaceLineEdit.setText('/network/lustre/iss02/')
      replaceLayout.addWidget(self._replaceLineEdit, alignment=Qt.AlignmentFlag.AlignCenter)
      replaceLayout.addWidget(QLabel('in all paths'), alignment=Qt.AlignmentFlag.AlignCenter)
      replaceLayout.addStretch()
      layout.addLayout(replaceLayout)
    else:
      parallelTrackingLayout = QHBoxLayout()
      self._parallelTrackingCheckbox = QCheckBox("Track multiple videos in parallel")
      self._parallelTrackingCheckbox.setChecked(True)
      self._parallelTrackingCheckbox.toggled.connect(lambda checked: self._processesLabel.setVisible(checked) or self._processesSpinBox.setVisible(checked))
      self._parallelTrackingCheckbox.setVisible(False)
      parallelTrackingLayout.addWidget(self._parallelTrackingCheckbox, alignment=Qt.AlignmentFlag.AlignCenter)
      self._processesLabel = QLabel('Number of processes:')
      self._processesLabel.setVisible(False)
      parallelTrackingLayout.addWidget(self._processesLabel, alignment=Qt.AlignmentFlag.AlignCenter)
      self._processesSpinBox = QSpinBox()
      self._processesSpinBox.setVisible(False)
      self._processesSpinBox.setValue(4)
      self._processesSpinBox.setRange(1, os.cpu_count())
      parallelTrackingLayout.addWidget(self._processesSpinBox, alignment=Qt.AlignmentFlag.AlignCenter)
      parallelTrackingLayout.addStretch()
      layout.addLayout(parallelTrackingLayout)

    folderPath = os.path.join(paths.getRootDataFolder(), 'trackingConfigurations')
    if not os.path.exists(folderPath):
      os.makedirs(folderPath)
    model = _VideoFilesModel()
    model.setRootPath(folderPath)
    self._tree = tree = _TrackingConfigurationsTreeView()
    tree.setModel(model)
    for idx in range(1, model.columnCount()):
      tree.hideColumn(idx)
    tree.setRootIndex(model.index(model.rootPath()))
    self._table = QTableView()
    selectionModel = _TrackingConfigurationsSelectionModel(app.window, self._table, model)
    tree.setSelectionModel(selectionModel)
    selectionModel.currentRowChanged.connect(lambda current, previous: current.row() == -1 or self._fileSelected(model.filePath(current)))

    treeLayout = QVBoxLayout()
    self._newConfigurationBtn = QPushButton("New configuration")
    self._newConfigurationBtn.clicked.connect(self._newConfiguration)
    treeLayout.addWidget(self._newConfigurationBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    treeLayout.addWidget(tree, stretch=1)
    treeWidget = QWidget()
    treeLayout.setContentsMargins(0, 0, 0, 0)
    treeWidget.setLayout(treeLayout)
    horizontalSplitter = util.CollapsibleSplitter()
    horizontalSplitter.addWidget(treeWidget)

    tableLayout = QVBoxLayout()
    tableButtonsLayout = QHBoxLayout()
    self._addVideosBtn = QPushButton("Add video(s)")
    self._addVideosBtn.clicked.connect(self._addVideos)
    self._addVideosBtn.clicked.connect(self._updateParallelTracking)
    tableButtonsLayout.addWidget(self._addVideosBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    self._addFolderBtn = QPushButton("Add folder")
    self._addFolderBtn.clicked.connect(self._addFolder)
    self._addFolderBtn.clicked.connect(self._updateParallelTracking)
    tableButtonsLayout.addWidget(self._addFolderBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    self._addMultipleFoldersBtn = QPushButton("Add multiple folders")
    self._addMultipleFoldersBtn.clicked.connect(self._addMultipleFolders)
    self._addMultipleFoldersBtn.clicked.connect(self._updateParallelTracking)
    tableButtonsLayout.addWidget(self._addMultipleFoldersBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    chooseConfigsBtn = QPushButton("Choose config for selected videos")
    chooseConfigsBtn.clicked.connect(lambda: self._table.model().setConfigs(sorted(set(map(lambda idx: idx.row(), self._table.selectionModel().selectedIndexes()))),
                                                                           QFileDialog.getOpenFileName(app.window, 'Select config file', paths.getConfigurationFolder(), "JSON (*.json)")[0]))
    chooseConfigsBtn.clicked.connect(self._updateParallelTracking)
    tableButtonsLayout.addWidget(chooseConfigsBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    removeVideosBtn = QPushButton("Remove selected videos")
    removeVideosBtn.clicked.connect(lambda: self._table.model().removeSelectedRows(sorted(set(map(lambda idx: idx.row(), self._table.selectionModel().selectedIndexes())))))
    removeVideosBtn.clicked.connect(self._updateParallelTracking)
    tableButtonsLayout.addWidget(removeVideosBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    saveChangesBtn = QPushButton("Save changes")
    saveChangesBtn.clicked.connect(lambda: self._table.model().saveFile(self._tree.model().filePath(selectionModel.currentIndex())) or QMessageBox.information(app.window, "Configuration saved", "Changes made to the configuration were saved."))
    tableButtonsLayout.addWidget(saveChangesBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    deleteConfigurationBtn = QPushButton("Delete configuration")
    deleteConfigurationBtn.clicked.connect(self._removeConfiguration)
    tableButtonsLayout.addWidget(deleteConfigurationBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    openConfigurationsFolderBtn = QPushButton("Open configurations folder")
    openConfigurationsFolderBtn.clicked.connect(lambda: QDesktopServices.openUrl(QUrl.fromLocalFile(folderPath)))
    tableButtonsLayout.addWidget(openConfigurationsFolderBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    self._runTrackingBtn = util.apply_style(QPushButton("Run tracking"), background_color=util.DEFAULT_BUTTON_COLOR)
    self._runTrackingBtn.clicked.connect(self._unsavedChangesWarning(lambda *_: self._runTracking(), forceSave=True))
    tableButtonsLayout.addWidget(self._runTrackingBtn, alignment=Qt.AlignmentFlag.AlignLeft)
    tableButtonsLayout.addStretch()
    tableLayout.addLayout(tableButtonsLayout)
    tableLayout.addWidget(self._table, stretch=1)

    self._mainWidget = QWidget()
    tableLayout.setContentsMargins(0, 0, 0, 0)
    self._mainWidget.setLayout(tableLayout)

    scrollArea = QScrollArea()
    scrollArea.setFrameShape(QFrame.Shape.NoFrame)
    scrollArea.setWidgetResizable(True)
    scrollArea.setWidget(self._mainWidget)
    horizontalSplitter.addWidget(scrollArea)
    horizontalSplitter.setStretchFactor(1, 1)
    horizontalSplitter.setChildrenCollapsible(False)
    layout.addWidget(horizontalSplitter, stretch=1)

    buttonsLayout = QHBoxLayout()
    buttonsLayout.addStretch()
    startPageBtn = QPushButton("Go to the start page")
    startPageBtn.clicked.connect(self._unsavedChangesWarning(lambda *_: app.show_frame("StartPage")))
    buttonsLayout.addWidget(startPageBtn, alignment=Qt.AlignmentFlag.AlignCenter)
    buttonsLayout.addStretch()
    layout.addLayout(buttonsLayout)

    self.setLayout(layout)
    self._mainWidget.hide()

  @contextlib.contextmanager
  def __selectAddedRows(self):
    app = QApplication.instance()
    model = self._table.model()
    firstNewRow = model.rowCount()
    yield
    if firstNewRow == model.rowCount():
      return  # no videos added
    addedRows = QItemSelection()
    addedRows.select(model.index(firstNewRow, 0), model.index(model.rowCount() - 1, 0))
    self._table.selectionModel().setCurrentIndex(model.index(firstNewRow, 0), QItemSelectionModel.SelectionFlag.ClearAndSelect | QItemSelectionModel.SelectionFlag.Rows)
    self._table.selectionModel().select(addedRows, QItemSelectionModel.SelectionFlag.ClearAndSelect | QItemSelectionModel.SelectionFlag.Rows)
    self._table.setFocus()

  def _addVideos(self):
    app = QApplication.instance()
    with self.__selectAddedRows():
      self._table.model().addVideos(QFileDialog.getOpenFileNames(app.window, 'Select one or more videos', os.path.expanduser("~"), "Video files (%s)" % ' '.join('*%s' % ext for ext in _VIDEO_EXTENSIONS))[0])

  def _addFolder(self):
    app = QApplication.instance()
    selectedFolder = QFileDialog.getExistingDirectory(app.window, "Select a folder", os.path.expanduser("~"))
    if selectedFolder is None:
      return
    with self.__selectAddedRows():
      return self._table.model().addVideos([os.path.normpath(os.path.join(root, f)) for root, _dirs, files in os.walk(selectedFolder) for f in files if os.path.splitext(f)[1] in _VIDEO_EXTENSIONS])

  def _addMultipleFolders(self):
    app = QApplication.instance()
    dialog = QFileDialog(app.window)
    dialog.setWindowTitle('Select one or more folders (use Ctrl or Shift key to select multiple)')
    dialog.setDirectory(os.path.expanduser("~"))
    dialog.setFileMode(QFileDialog.FileMode.Directory)
    dialog.setOption(QFileDialog.Option.DontUseNativeDialog, True)
    dialog.setOption(QFileDialog.Option.ShowDirsOnly, True)
    listView = dialog.findChild(QListView, 'listView')
    if listView:
      listView.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
    treeView = dialog.findChild(QTreeView)
    if treeView:
      treeView.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

    if not dialog.exec():
      return None
    with self.__selectAddedRows():
      self._table.model().addVideos([os.path.normpath(os.path.join(root, f)) for path in dialog.selectedFiles()
                                     for root, _dirs, files in os.walk(path) for f in files if os.path.splitext(f)[1] in _VIDEO_EXTENSIONS])

  def _updateParallelTracking(self):
    if self._ZZkwargs.get('sbatchMode', False) or self._ZZkwargs.get('findMultipleROIs', False) or self._ZZkwargs.get('headEmbedded', False):
      return
    videos, configs = self._table.model().getData()
    enabled = bool(configs)
    for config in configs:
      if not os.path.exists(config):
        enabled = False
        break
      with open(config) as f:
        cfg = json.load(f)
      if not (cfg.get('fasterMultiprocessing', False) or cfg.get('headEmbeded', False) or (cfg.get('trackingImplementation') == "fastFishTracking.tracking")):
        enabled = False
        break
    self._parallelTrackingCheckbox.setVisible(enabled)
    self._processesLabel.setVisible(enabled and self._parallelTrackingCheckbox.isChecked())
    self._processesSpinBox.setVisible(enabled and self._parallelTrackingCheckbox.isChecked())

  def _unsavedChangesWarning(self, fn, forceSave=False):
    app = QApplication.instance()
    def inner(*args, **kwargs):
      if forceSave:
        text = "Do you want to save the changes and proceed?"
      else:
        text = "Are you sure you want to proceed? Unsaved changes will be lost."
      filename = self._tree.model().filePath(self._tree.selectionModel().currentIndex())
      if self._table.model() is not None and self._table.model().hasUnsavedChanges(filename):
        if QMessageBox.question(app.window, "Unsaved changes", text, defaultButton=QMessageBox.StandardButton.No) != QMessageBox.StandardButton.Yes:
          return
        elif forceSave:
          self._table.model().saveFile(filename)
      return fn(*args, **kwargs)
    return inner

  def _newConfiguration(self):
    number = 1
    while os.path.exists(os.path.join(self._tree.model().rootPath(), 'Configuration %d.csv' % number)):
      number += 1
    path = os.path.join(self._tree.model().rootPath(), 'Configuration %d.csv' % number)
    pd.DataFrame(columns=_VideosModel._COLUMN_TITLES).to_csv(path, index=False)
    index = self._tree.model().index(path)
    self._tree.selectionModel().setCurrentIndex(index, QItemSelectionModel.SelectionFlag.ClearAndSelect)
    self._tree.edit(index)

  def _removeConfiguration(self):
    app = QApplication.instance()
    if QMessageBox.question(app.window, "Delete configuration", "Are you sure you want to delete the configuration? This action removes the file from disk and cannot be undone.",
                            defaultButton=QMessageBox.StandardButton.No) != QMessageBox.StandardButton.Yes:
      return
    pathToRemove = self._tree.model().filePath(self._tree.selectionModel().currentIndex())
    self._tree.model().remove(self._tree.selectionModel().currentIndex())
    if pathToRemove == self._tree.model().filePath(self._tree.selectionModel().currentIndex()):  # last valid file removed
      self._mainWidget.hide()

  def _fileSelected(self, filename):
    self._mainWidget.show()
    self._table.setModel(_VideosModel(filename))
    self._table.horizontalHeader().resizeSections(QHeaderView.ResizeMode.Stretch)
    self._updateParallelTracking()

  def _runTracking(self):
    app = QApplication.instance()
    videos, configs = self._table.model().getData()
    errors = []
    for video, config in zip(videos, configs):
      if not config:
        errors.append('Config is not specified for video %s.' % video)
      elif not os.path.exists(config):
        errors.append('Config %s does not exist.' % config)
      if not os.path.exists(video):
        errors.append('Video %s does not exist.' % video)
    if errors:
      error = QMessageBox(app.window)
      error.setIcon(QMessageBox.Icon.Critical)
      error.setWindowTitle("Specification contains some errors")
      error.setText("Cannot run tracking becase the specified video and config combinations contain some errors:")
      error.setDetailedText("\n".join(errors))
      textEdit = error.findChild(QTextEdit)
      textEdit.setLineWrapMode(QTextEdit.LineWrapMode.NoWrap)
      layout = error.layout()
      layout.addItem(QSpacerItem(600, 0), layout.rowCount(), 0, 1, layout.columnCount())
      error.exec()
      return
    if self._ZZkwargs.get('sbatchMode', False):
      videos = [video.replace('\\', '/').replace(self._originalLineEdit.text(), self._replaceLineEdit.text()) for video in videos]
    elif self._processesSpinBox.isVisible():
      self._ZZkwargs['processes'] = self._processesSpinBox.value()
    defineWellsVideos = []
    headEmbeddedVideos = []
    for video, config in zip(videos, configs):
      if self._ZZkwargs.get('findMultipleROIs', False) or self._ZZkwargs.get('headEmbedded', False):
        break
      with open(config) as f:
        cfg = json.load(f)
      videoName = os.path.splitext(os.path.basename(video))[0]
      inputsFolder = os.path.join(app.ZZoutputLocation, '.ZebraZoomVideoInputs', videoName)
      if (cfg.get("multipleROIsDefinedDuringExecution", False) or cfg.get("groupOfMultipleSameSizeAndShapeEquallySpacedWells", False)) and \
          not os.path.exists(os.path.join(inputsFolder, 'intermediaryWellPositionReloadNoMatterWhat.txt')):
        defineWellsVideos.append((video, config))
      if cfg.get("headEmbeded", False) and (not os.path.exists(os.path.join(inputsFolder, f'{videoName}.csv')) or not os.path.exists(os.path.join(inputsFolder, f'{videoName}HP.csv'))):
        headEmbeddedVideos.append((video, config))
    if defineWellsVideos:
      text = 'Some of the videos do not have the regions defined. Would you like to define them before running tracking?'
      sameCoordinatesCheckbox = QCheckBox("Use the same coordinates for all videos")
      msgbox = QMessageBox(QMessageBox.Icon.Question, "Missing required coordinates", text, parent=app.window,
                           buttons=QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
      msgbox.setDefaultButton(QMessageBox.StandardButton.Yes)
      msgbox.setCheckBox(sameCoordinatesCheckbox)
      if msgbox.exec() == QMessageBox.StandardButton.Yes:
        ZZkwargs = self._ZZkwargs.copy()
        ZZkwargs.update({'findMultipleROIs': True, 'askCoordinatesForAll': not sameCoordinatesCheckbox.isChecked(), 'processes': 1})
        launchZebraZoom(*zip(*defineWellsVideos), **ZZkwargs)
    if headEmbeddedVideos:
      text = 'Some of the head-embedded videos do not have the required head and tail coordinates defined. Would you like to define them before running tracking?'
      msgbox = QMessageBox(QMessageBox.Icon.Question, "Missing required coordinates", text, parent=app.window,
                           buttons=QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
      msgbox.setDefaultButton(QMessageBox.StandardButton.Yes)
      if msgbox.exec() == QMessageBox.StandardButton.Yes:
        ZZkwargs = self._ZZkwargs.copy()
        ZZkwargs.update({'headEmbedded': True, 'processes': 1})
        launchZebraZoom(*zip(*headEmbeddedVideos), **ZZkwargs)
    app.show_frame("Patience")
    kwargs = self._ZZkwargs.copy()
    kwargs['showValidationVideoWidget'] = not (kwargs.get('sbatchMode', False) or kwargs.get('findMultipleROIs', False) or kwargs.get('headEmbedded', False))
    app.window.centralWidget().layout().currentWidget().setArgs((videos, configs), kwargs)


def _showTemporaryPage(ZZkwargs):
  app = QApplication.instance()
  layout = app.window.centralWidget().layout()
  page = _VideoSelectionPage(ZZkwargs)
  layout.addWidget(page)
  layout.setCurrentWidget(page)
  def cleanup():
    layout.removeWidget(page)
    layout.currentChanged.disconnect(cleanup)
  layout.currentChanged.connect(cleanup)


def chooseFolderToAnalyze(self, sbatchMode):
  ZZkwargs = {'sbatchMode': sbatchMode}
  _showTemporaryPage(ZZkwargs)


def chooseFolderForTailExtremityHE(self):
  ZZkwargs = {'headEmbedded': True}
  _showTemporaryPage(ZZkwargs)


def chooseFolderForMultipleROIs(self, askCoordinatesForAll):
  ZZkwargs = {'findMultipleROIs': True, 'askCoordinatesForAll': askCoordinatesForAll}
  _showTemporaryPage(ZZkwargs)


def chooseConfigFile(ZZargs, ZZkwargs):
  app = QApplication.instance()
  configFileName, _ = QFileDialog.getOpenFileName(app.window, 'Select file', paths.getConfigurationFolder(), "JSON (*.json)")
  if not configFileName:
    return
  ZZargs += ([configFileName],)
  if globalVariables["mac"] or globalVariables["lin"]:
    app.show_frame("Patience")
    app.window.centralWidget().layout().currentWidget().setArgs(ZZargs, ZZkwargs)
  else:
    launchZebraZoom(*ZZargs, **ZZkwargs)


def _runTracking(args, createValidationVideo, ZZoutputLocation):
  videoPath, config = args
  path        = os.path.dirname(videoPath)
  nameWithExt = os.path.basename(videoPath)
  name        = os.path.splitext(nameWithExt)[0]
  videoExt    = os.path.splitext(nameWithExt)[1][1:]

  tabParams = ["zebraZoomVideoAnalysis", path, name, videoExt, config, "freqAlgoPosFollow", 100, "outputFolder", ZZoutputLocation]
  if createValidationVideo is not None:
    tabParams.extend(['createValidationVideo', createValidationVideo])
  try:
    ZebraZoomVideoAnalysis(path, name, videoExt, config, tabParams).run()
  except ValueError:
    print("moving on to the next video for ROIs identification")
  except NameError:
    return


@util.showInProgressPage('Tracking')
def launchZebraZoom(videos, configs, headEmbedded=False, sbatchMode=False, createValidationVideo=None, findMultipleROIs=False,
                    askCoordinatesForAll=True, firstFrame=None, lastFrame=None, backgroundExtractionForceUseAllVideoFrames=None, processes=1):
  app = QApplication.instance()

  if sbatchMode:
    commandsFile = open(os.path.join(paths.getRootDataFolder(), "commands.txt"), "w", newline='\n')
    nbVideosToLaunch = 0

  if len(videos) > 1:
    if not askCoordinatesForAll:
      videosGenerator = iter(videos)
      videos = [next(videosGenerator)]
  else:
    videoPath = videos[0]
    videoName = os.path.splitext(os.path.basename(videoPath))[0]
    inputsFolder = os.path.join(app.ZZoutputLocation, '.ZebraZoomVideoInputs', videoName)
    if (os.path.exists(os.path.join(inputsFolder, f'{videoName}HP.csv')) or os.path.exists(os.path.join(inputsFolder, f'{videoName}.csv'))) and \
        QMessageBox.question(app.window, "Previously stored coordinates found", "Do you want to use the previously stored coordinates?",
                             defaultButton=QMessageBox.StandardButton.Yes) != QMessageBox.StandardButton.Yes:
      if os.path.exists(os.path.join(inputsFolder, f'{videoName}HP.csv')):
        os.remove(os.path.join(inputsFolder, f'{videoName}HP.csv'))
      if os.path.exists(os.path.join(inputsFolder, f'{videoName}.csv')):
        os.remove(os.path.join(inputsFolder, f'{videoName}.csv'))

  print("allVideos:", videos)

  if processes > 1 and len(videos) > 1 and not sbatchMode:
    with Pool(min(processes, len(videos))) as pool:
      pool.map(partial(_runTracking, createValidationVideo=createValidationVideo, ZZoutputLocation=app.ZZoutputLocation), zip(videos, configs))
  else:
    for videoPath, config in zip(videos, configs):

      path        = os.path.dirname(videoPath)
      nameWithExt = os.path.basename(videoPath)
      name        = os.path.splitext(nameWithExt)[0]
      videoExt    = os.path.splitext(nameWithExt)[1][1:]

      if not headEmbedded:
        if len(videos) == 1:
          tabParams = ["zebraZoomVideoAnalysis", path, name, videoExt, config, "freqAlgoPosFollow", 100, "popUpAlgoFollow", 1, "outputFolder", app.ZZoutputLocation]
        else:
          tabParams = ["zebraZoomVideoAnalysis", path, name, videoExt, config, "freqAlgoPosFollow", 100, "outputFolder", app.ZZoutputLocation]
        if backgroundExtractionForceUseAllVideoFrames is not None:
          tabParams.extend(["backgroundExtractionForceUseAllVideoFrames", int(backgroundExtractionForceUseAllVideoFrames)])
        if firstFrame is not None:
          tabParams.extend(["firstFrame", firstFrame])
        if lastFrame is not None:
          tabParams.extend(["lastFrame", lastFrame])
        if createValidationVideo is not None:
          tabParams.extend(['createValidationVideo', createValidationVideo])
        if findMultipleROIs:
          tabParams = tabParams + ["exitAfterWellsDetection", 1, "saveWellPositionsToBeReloadedNoMatterWhat", 1]
        try:
          if sbatchMode:
            commandsFile.write('python -m zebrazoom ' + ' '.join(tabParams[1:4]) + ' configFiles/%s\n' % os.path.basename(config))
            nbVideosToLaunch = nbVideosToLaunch + 1
          else:
            ZebraZoomVideoAnalysis(path, name, videoExt, config, tabParams).run()
        except ValueError:
          print("moving on to the next video for ROIs identification")

      else:
        tabParams = ["outputFolder", app.ZZoutputLocation]
        if backgroundExtractionForceUseAllVideoFrames is not None:
          tabParams.extend(["backgroundExtractionForceUseAllVideoFrames", int(backgroundExtractionForceUseAllVideoFrames)])
        if firstFrame is not None:
          tabParams.extend(["firstFrame", firstFrame])
        if lastFrame is not None:
          tabParams.extend(["lastFrame", lastFrame])
        getTailExtremityFirstFrame(path, name, videoExt, config, tabParams)

  if len(videos) > 1 and findMultipleROIs and not askCoordinatesForAll:
    videoInputsFolder = os.path.join(app.ZZoutputLocation, '.ZebraZoomVideoInputs', os.path.splitext(os.path.basename(videos[0]))[0])
    coordinatesFile = os.path.join(videoInputsFolder, 'intermediaryWellPositionReloadNoMatterWhat.txt')
    rotationFile = os.path.join(videoInputsFolder, 'rotationAngle.txt')
    for video in videosGenerator:
      folderPath = os.path.join(app.ZZoutputLocation, '.ZebraZoomVideoInputs', os.path.splitext(os.path.basename(video))[0])
      if not os.path.exists(folderPath):
        os.makedirs(folderPath)
      shutil.copy2(coordinatesFile, os.path.join(folderPath, 'intermediaryWellPositionReloadNoMatterWhat.txt'))
      if os.path.exists(rotationFile):
        shutil.copy2(rotationFile, os.path.join(folderPath, 'rotationAngle.txt'))

  if sbatchMode:

    commandsFile.close()

    with open(configs[0]) as f:
      jsonFile = json.load(f)
    nbWells = jsonFile["nbWells"]
    if nbWells > 24:
      nbWells = 24

    launchFile = open(os.path.join(paths.getRootDataFolder(), "launchZZ.sh"), "w", newline='\n')
    linesToWrite = ['#!/bin/sh',
                    '#SBATCH --ntasks=1',
                    '#SBATCH --cpus-per-task='+str(nbWells),
                    '#SBATCH --array=1-'+str(nbVideosToLaunch),
                    '#SBATCH --mem=16G',
                    '#SBATCH --time=23:00:00',
                    '#SBATCH --partition=normal',
                    '#SBATCH --job-name="ZebraZoom-protocole"',
                    '',
                    'module load python/3.8',
                    'source activate zebrazoom',
                    '',
                    'date',
                    '',
                    'export CMD_FILE_PATH=./commands.txt',
                    '',
                    'export CMD=$(sed -n ${SLURM_ARRAY_TASK_ID}p ${CMD_FILE_PATH})',
                    '',
                    'echo $CMD',
                    'eval $CMD',
                    '',
                    'date']
    linesToWrite = [line + '\n' for line in linesToWrite]
    launchFile.writelines(linesToWrite)
    launchFile.close()

    configsFolder = os.path.join(paths.getRootDataFolder(), 'configFiles')
    if os.path.exists(configsFolder):
      shutil.rmtree(configsFolder)
    os.makedirs(configsFolder)
    for config in configs:
      shutil.copy2(config, configsFolder)

    app.show_frame("ZZoutroSbatch")

  else:

    app.show_frame("ZZoutro")


def openConfigurationFileFolder(self, homeDirectory):
  dir_path = os.path.join(homeDirectory,'configuration')
  if sys.platform == "win32":
    os.startfile(dir_path)
  else:
    opener ="open" if sys.platform == "darwin" else "xdg-open"
    subprocess.call([opener, dir_path])


def openZZOutputFolder(self, homeDirectory):
  if len(self.ZZoutputLocation):
    dir_path = self.ZZoutputLocation
  else:
    dir_path = os.path.join(homeDirectory,'ZZoutput')
  if sys.platform == "win32":
    os.startfile(dir_path)
  else:
    opener ="open" if sys.platform == "darwin" else "xdg-open"
    subprocess.call([opener, dir_path])



import os
import sys


def getRootDataFolder():
  if getattr(sys, 'frozen', False):
    installationFolder = os.path.dirname(sys.executable)
    if os.path.exists(os.path.join(installationFolder, 'Uninstall.exe')):  # installed using the installer
      from PyQt5.QtCore import QStandardPaths
      return os.path.join(QStandardPaths.writableLocation(QStandardPaths.StandardLocation.DocumentsLocation), 'ZebraZoom')
    return os.path.join(installationFolder, 'zebrazoom')  # simply extracted zip
  return os.path.dirname(os.path.dirname(os.path.realpath(__file__)))  # running in a normal Python environment


def getDefaultZZoutputFolder():
  return os.path.join(getRootDataFolder(), 'ZZoutput')


def getConfigurationFolder():
  return os.path.join(getRootDataFolder(), 'configuration')


def getRolloverDetectionConfigurationFolder():
  return os.path.join(getRootDataFolder(), 'rolloverDetectionConfiguration')


def getDataAnalysisFolder():
  return os.path.join(getRootDataFolder(), 'dataAnalysis')



import numpy as np
import queue
import math
import cv2

def getListOfWellsOnWhichToRunTheTracking(self, frame1, frame2):
  
  listOfWellsOnWhichToRunTheTracking = []
  
  for wellNumber in range(0, len(self._wellPositions)):
  
    # Retrieving well/tank/arena coordinates and selecting ROI
    wellXtop = self._wellPositions[wellNumber]['topLeftX']
    wellYtop = self._wellPositions[wellNumber]['topLeftY']
    lenghtWell_X = self._wellPositions[wellNumber]['lengthX']
    lenghtWell_Y = self._wellPositions[wellNumber]['lengthY']
    frameROI1 = frame1[wellYtop:wellYtop+lenghtWell_Y, wellXtop:wellXtop+lenghtWell_X]
    frameROI2 = frame2[wellYtop:wellYtop+lenghtWell_Y, wellXtop:wellXtop+lenghtWell_X]
    
    res = cv2.absdiff(frameROI1, frameROI2)
    ret, res = cv2.threshold(res,self._hyperparameters["thresForDetectMovementWithRawVideo"],255,cv2.THRESH_BINARY)
    totDiff = cv2.countNonZero(res)
    
    if totDiff > self._hyperparameters["minNbPixelForDetectMovementInWell"]:
      listOfWellsOnWhichToRunTheTracking.append(wellNumber)
    
  return listOfWellsOnWhichToRunTheTracking



import importlib
import os
import pkgutil


def _init():
  dirname = os.path.dirname(__file__)
  for dir_ in (os.path.join(dirname, name) for name in os.listdir(dirname) if os.path.isdir(os.path.join(dirname, name))):
    for loader, module, is_pkg in pkgutil.iter_modules([dir_], prefix=f'{__name__}.{os.path.basename(dir_)}.'):
      loader.find_module(module).load_module(module)


_init()
del _init



import numpy as np

from ._openResultsFile import openResultsFile


def getDataPerTimeInterval(videoName: str, numWell: int, numAnimal: int, startTimeInSeconds: float, endTimeInSeconds: float, parameterName: str) -> np.array:
  if startTimeInSeconds >= endTimeInSeconds:
    raise ValueError('end time must be larger than start time')
  with openResultsFile(videoName, 'r') as results:
    firstFrame = results.attrs['firstFrame']
    firstFrameInSeconds = firstFrame / results.attrs['videoFPS']
    lastFrameInSeconds = results.attrs['lastFrame'] / results.attrs['videoFPS']
    if startTimeInSeconds < firstFrameInSeconds or endTimeInSeconds > lastFrameInSeconds:
      raise ValueError(f'Tracking was performed from {firstFrameInSeconds}s to {lastFrameInSeconds}s, start and end times must be within this interval.')
    if 'videoFPS' not in results.attrs:
      raise ValueError(f'videoFPS not found in the results, cannot convert seconds to frames')
    intervalStart = int(startTimeInSeconds * results.attrs['videoFPS']) - firstFrame
    intervalEnd = int(endTimeInSeconds * results.attrs['videoFPS']) - firstFrame
    dataGroup = results[f'dataForWell{numWell}/dataForAnimal{numAnimal}/dataPerFrame']
    if parameterName not in dataGroup:
      allParameters = ', '.join(dataGroup.keys())
      raise ValueError(f'Parameter "{parameterName}" not found in results. Available parameters: {allParameters}')
    dataset = dataGroup[parameterName]
    return np.array(dataset[intervalStart:intervalEnd]) if 'columns' not in dataset.attrs else np.column_stack([dataset[col][intervalStart:intervalEnd] for col in dataset.attrs['columns']])



import numpy as np

from ._calculateAndStoreTailAngleHeatmap import calculateAndStoreTailAngleHeatmap
from ._openResultsFile import openResultsFile


def getTailAngleHeatmapPerTimeInterval(videoName: str, numWell: int, numAnimal: int, startTimeInSeconds: float, endTimeInSeconds: float) -> tuple:
  if startTimeInSeconds >= endTimeInSeconds:
    raise ValueError('end time must be larger than start time')
  with openResultsFile(videoName, 'r+') as results:
    firstFrame = results.attrs['firstFrame']
    firstFrameInSeconds = firstFrame / results.attrs['videoFPS']
    lastFrameInSeconds = results.attrs['lastFrame'] / results.attrs['videoFPS']
    if startTimeInSeconds < firstFrameInSeconds or endTimeInSeconds > lastFrameInSeconds:
      raise ValueError(f'Tracking was performed from {firstFrameInSeconds}s to {lastFrameInSeconds}s, start and end times must be within this interval.')
    if 'videoFPS' not in results.attrs:
      raise ValueError(f'videoFPS not found in the results, cannot convert seconds to frames')
    boutsPath = f'dataForWell{numWell}/dataForAnimal{numAnimal}/listOfBouts'
    if boutsPath not in results:
      raise ValueError(f"bouts not found for animal {numAnimal} in well {numWell}")
    boutsGroup = results[boutsPath]
    intervalStart = int(startTimeInSeconds * results.attrs['videoFPS']) - firstFrame
    intervalEnd = int(endTimeInSeconds * results.attrs['videoFPS']) - firstFrame
    dataGroup = results[f'dataForWell{numWell}/dataForAnimal{numAnimal}/dataPerFrame']
    if 'tailAngleHeatmap' in dataGroup:
      tailAngleHeatmap = dataGroup['tailAngleHeatmap']
    else:
      print(f'calculating and storing tail angle heatmap for all bouts for well {numWell}, animal {numAnimal}')
      tailAngleHeatmap = calculateAndStoreTailAngleHeatmap(results, dataGroup, boutsGroup)
    return [tailAngleHeatmap[col][intervalStart:intervalEnd] for col in tailAngleHeatmap.dtype.names], int(startTimeInSeconds * results.attrs['videoFPS']), dataGroup['TailLength'][0]



from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.detectMovementWithRawVideoInsideTracking import detectMovementWithRawVideoInsideTracking
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.trackTail import trackTail
# import zebrazoom.code.tracking
import numpy as np
import time
import math
import cv2

def backgroundSubtractionOnlyOnROIs(self, frame, k):

  # Color to grey scale transformation
  frame = frame[:,:,0]
  
  # Bout detection
  if self._hyperparameters["detectMovementWithRawVideoInsideTracking"]:
    detectMovementWithRawVideoInsideTracking(self, k, frame)
  
  # Going through each well/arena/tank and applying tracking method on it
  t1 = time.time()
  for wellNumber in self._listOfWellsOnWhichToRunTheTracking:
    
    if self._hyperparameters["detectMovementWithRawVideoInsideTracking"] == 0 or k <= 2 or np.sum([self._auDessusPerAnimalIdList[wellNumber][i][k] for i in range(0, self._hyperparameters["nbAnimalsPerWell"])]):
      
      for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        
        if self._hyperparameters["detectMovementWithRawVideoInsideTracking"] == 0 or self._auDessusPerAnimalIdList[wellNumber][animalId][k] or k <= 2:
          
          # Retrieving ROI coordinates and selecting ROI
          roiXStart = self._wellPositions[wellNumber]['topLeftX'] + int(self._trackingDataPerWell[wellNumber][animalId][k-1][0][0] - self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"])
          roiYStart = self._wellPositions[wellNumber]['topLeftY'] + int(self._trackingDataPerWell[wellNumber][animalId][k-1][0][1] - self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"])
          roiXEnd   = roiXStart + 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
          roiYEnd   = roiYStart + 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
          if roiXStart < 0:
            roiXStart = 0
            roiXEnd   = 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
          if roiYStart < 0:
            roiYStart = 0
            roiYEnd   = 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
          if roiXEnd >= len(frame[0]):
            roiXEnd   = len(frame[0]) - 1
            roiXStart = len(frame[0]) - 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
          if roiYEnd >= len(frame):
            roiYEnd   = len(frame) - 1
            roiYStart = len(frame) - 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
          frameROI = frame[roiYStart:roiYEnd, roiXStart:roiXEnd].copy()
          
          # Subtracting background of image
          backgroundROI = self._background[roiYStart:roiYEnd, roiXStart:roiXEnd]
          frameROI = 255 - np.where(backgroundROI >= frameROI, backgroundROI - frameROI, 0).astype(np.uint8)

          # Applying gaussian filter
          paramGaussianBlur = self._hyperparameters["paramGaussianBlur"]
          frameROI = cv2.GaussianBlur(frameROI, (paramGaussianBlur, paramGaussianBlur), 0)
          
          # Head position tracking
          (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(frameROI)
          if minVal >= self._hyperparameters["minimumHeadPixelValue"]:
            self._trackingDataPerWell[wellNumber][animalId][k] = self._trackingDataPerWell[wellNumber][animalId][k-1]
          else:
            if self._hyperparameters["trackTail"]:
              # Tail tracking
              a, self._lastFirstTheta[wellNumber] = trackTail(self, frameROI, headPosition, self._hyperparameters, wellNumber, k, self._lastFirstTheta[wellNumber])
              a[0, :, 0] += roiXStart - self._wellPositions[wellNumber]['topLeftX']
              a[0, :, 1] += roiYStart - self._wellPositions[wellNumber]['topLeftY']
              self._trackingDataPerWell[wellNumber][animalId][k][:len(a[0])] = a
            else:
              self._trackingDataPerWell[wellNumber][animalId][k][0][0] = headPosition[0] + roiXStart - self._wellPositions[wellNumber]['topLeftX']
              self._trackingDataPerWell[wellNumber][animalId][k][0][1] = headPosition[1] + roiYStart - self._wellPositions[wellNumber]['topLeftY']
        else:
          if k > 0:
            self._trackingDataPerWell[wellNumber][animalId][k] = self._trackingDataPerWell[wellNumber][animalId][k-1]
        
        # 'Removing' animal just tracked
        if self._hyperparameters["nbAnimalsPerWell"] > 1:
          frame = cv2.circle(frame.copy(), (int(self._wellPositions[wellNumber]['topLeftX'] + self._trackingDataPerWell[wellNumber][animalId][k][0][0]), int(self._wellPositions[wellNumber]['topLeftY'] + self._trackingDataPerWell[wellNumber][animalId][k][0][1])), int(self._hyperparameters["maxDepth"]/4), (255, 255, 255), -1)
          for pointOnTail in range(1, len(self._trackingDataPerWell[wellNumber][animalId][k])):
            start_point = (int(self._wellPositions[wellNumber]['topLeftX'] + self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail-1][0]), int(self._wellPositions[wellNumber]['topLeftY'] + self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail-1][1]))
            end_point   = (int(self._wellPositions[wellNumber]['topLeftX'] + self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail][0]), int(self._wellPositions[wellNumber]['topLeftY'] + self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail][1]))
            cv2.line(frame, start_point, end_point, (255, 255, 255), int(self._hyperparameters["maxDepth"]/5))
        
      # Id invertion if necessary
      if self._hyperparameters["nbAnimalsPerWell"] > 1:
        # NEED TO IMPROVE THIS IN THE FUTURE!!!
        for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]-1):
          dist_animalIdCurFrame_animal0PrevFrame = math.sqrt((self._trackingDataPerWell[wellNumber][animalId][k][0][0] - self._trackingDataPerWell[wellNumber][0][k-1][0][0])**2     +     (self._trackingDataPerWell[wellNumber][animalId][k][0][1] - self._trackingDataPerWell[wellNumber][0][k-1][0][1])**2)
          dist_animalIdCurFrame_animal1PrevFrame = math.sqrt((self._trackingDataPerWell[wellNumber][animalId][k][0][0] - self._trackingDataPerWell[wellNumber][1][k-1][0][0])**2     +     (self._trackingDataPerWell[wellNumber][animalId][k][0][1] - self._trackingDataPerWell[wellNumber][1][k-1][0][1])**2)
          if dist_animalIdCurFrame_animal1PrevFrame < dist_animalIdCurFrame_animal0PrevFrame:
            temp = self._trackingDataPerWell[wellNumber][0][k].copy()
            self._trackingDataPerWell[wellNumber][0][k] = self._trackingDataPerWell[wellNumber][1][k]
            self._trackingDataPerWell[wellNumber][1][k] = temp
    else:  
      for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        self._trackingDataPerWell[wellNumber][animalId][k] = self._trackingDataPerWell[wellNumber][animalId][k-1]


import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import cv2
import os

from zebrazoom.dataAnalysis.dataanalysis import sortGenotypes

def visualizeClusters(dfParam, classifications, predictedProbas, modelUsedForClustering, nbConditions, nbCluster, nbFramesTakenIntoAccount, scaleGraphs, showFigures, outputFolderResult, videoSaveFirstTenBouts, globalParametersCalculations):
  
  instaTBF   = ['instaTBF'+str(i)  for i in range(1, nbFramesTakenIntoAccount + 1)]
  instaAmp   = ['instaAmp'+str(i)  for i in range(1, nbFramesTakenIntoAccount + 1)]
  instaAsym  = ['instaAsym'+str(i) for i in range(1, nbFramesTakenIntoAccount + 1)]
  tailAngles = ['tailAngles'+str(i) for i in range(1, nbFramesTakenIntoAccount + 1)]
  
  instaSpeed       = ['instaSpeed'       + str(i) for i in range(1, nbFramesTakenIntoAccount + 1)]
  instaHeadingDiff = ['instaHeadingDiff' + str(i) for i in range(1, nbFramesTakenIntoAccount + 1)]
  instaHorizDispl  = ['instaHorizDispl'  + str(i) for i in range(1, nbFramesTakenIntoAccount + 1)]
  
  possibleColors = ['b', 'r', 'g', 'k', 'c', 'm', 'y']
  possibleColorsNames = ['blue', 'red', 'green', 'black', 'cyan', 'magenta', 'yellow']
  
  xaxis    = [0, 30]
  freqAxis = [0, 0.5]
  ampAxis  = [0, 1.6]
  asymAxis = [0, 0.8]
  angAxis  = [0, 1.5]
  
  dfParam['classification'] = classifications
  
  # if modelUsedForClustering == 'GaussianMixture':
    # for j in range(0, nbLabels):
      # probasClassJ = predictedProbas[:, sortedIndices[j]]
      # dfParam['classProba' + str(j)] = probasClassJ
  
  # Calculating proportions of each conditions in each class
  for clusterProportionsPerFish in [True, False]:
    proportions = np.zeros((nbConditions, nbCluster))
    if clusterProportionsPerFish:
      df2 = dfParam[['Trial_ID', 'Well_ID', 'Condition', 'classification']]
      for i in range(0, nbCluster):
        df2.loc[df2['classification'] == i, 'classifiedAs'+str(i)] = 1
        df2.loc[df2['classification'] != i, 'classifiedAs'+str(i)] = 0
      df2 = df2.groupby(['Trial_ID', 'Well_ID', 'Condition']).sum()
      df2['Condition'] = [elem[2] for elem in df2.index]
      df2['totalNbOfBouts'] = df2['classifiedAs0']
      if nbCluster >= 2:
        for i in range(1, nbCluster):
          df2['totalNbOfBouts'] = df2['totalNbOfBouts'] + df2['classifiedAs' + str(i)]
      df2['totalNbOfBouts'] = df2['totalNbOfBouts'].replace(0, 1)
      for i in range(0, nbCluster):
        df2['classifiedAs' + str(i)] = df2['classifiedAs' + str(i)] / df2['totalNbOfBouts']
      
      df2.to_excel(os.path.join(outputFolderResult, 'clusterProportionsPerAnimal.xlsx'))
      
      for idxCond, cond in enumerate(np.unique(dfParam['Condition'].values)):
        for classed in range(0, len(proportions[0])):
          proportions[idxCond, classed] = np.median(df2.loc[df2['Condition'] == cond]['classifiedAs' + str(classed)])
      
      conditions = list(sortGenotypes(dfParam["Condition"].unique().tolist()))
      fig, tabAx = plt.subplots(1, len(proportions[0]), figsize=(22.9, 8.8))
      for classed in range(0, len(proportions[0])):
        b = sns.boxplot(ax=tabAx[int(classed)], data=df2, x='Condition', y='classifiedAs' + str(classed), showmeans=1, showfliers=1, order=conditions)
        c = sns.stripplot(ax=tabAx[int(classed)], data=df2, x='Condition', y='classifiedAs' + str(classed), color='red', size=7, order=conditions)
        b.set_ylabel('', fontsize=0)
        b.set_xlabel('', fontsize=0)
        b.axes.set_title('Cluster ' + str(classed), fontsize=30)
      plt.savefig(os.path.join(outputFolderResult, 'clustersProportionsPerFish.png'))
      
    else:
      
      df2 = dfParam[['Condition','classification']]
      for idxCond, cond in enumerate(np.unique(dfParam['Condition'].values)):
        for classed in range(0, len(proportions[0])):
          proportions[idxCond, classed] = len(df2.loc[(df2['Condition'] == cond) & (df2['classification'] == classed)])
      for i in range(0, nbConditions):
        proportions[i, :] = proportions[i, :] / sum(proportions[i, :])
    
    if clusterProportionsPerFish:
      outF = open(os.path.join(outputFolderResult, 'proportionsPerFish.txt'), "w")
    else:
      outF = open(os.path.join(outputFolderResult, 'proportionsPerBout.txt'), "w")
    
    labelX = ""
    for i in range(0, nbCluster):
      labelX = labelX + "Cluster " + str(i+1) + " : \n"
      for j, cond in enumerate(np.unique(dfParam['Condition'].values)):
        labelX = labelX + str(cond) + ": " + str(round(proportions[j,i]*100*100)/100) + "%, "
        labelX = labelX + "\n"
      labelX = labelX + "\n"
    outF.write(labelX)
    outF.write("\n")
    outF.close()
  
  # Plotting each cluster one by one
  
  mostRepresentativeBout = np.zeros((nbConditions,nbCluster))
  
  fig, tabAx = plt.subplots(4, len(proportions[0]), figsize=(22.9, 8.8))
  
  conditions = list(sortGenotypes(dfParam["Condition"].unique().tolist()))
  for idxCond, cond in enumerate(conditions):
    for classed in range(0, len(proportions[0])):
      dfTemp = dfParam.loc[(dfParam['Condition'] == cond) & (dfParam['classification'] == classed)]
      instaTBFtab   = dfTemp[instaTBF]
      instaAmptab   = dfTemp[instaAmp]
      instaAsymtab  = dfTemp[instaAsym]
      tailAnglestab = dfTemp[tailAngles]
      color = possibleColors[idxCond]
      if nbCluster == 1:
        tabAx[0].plot(instaTBFtab.median().values,  color, label=cond)
        tabAx[1].plot(instaAmptab.median().values,  color)
        tabAx[2].plot(instaAsymtab.median().values, color)
        tabAx[3].plot(tailAnglestab.median().values, color)
      else:
        tabAx[0, classed].plot(instaTBFtab.median().values,  color, label=cond)
        tabAx[1, classed].plot(instaAmptab.median().values,  color)
        tabAx[2, classed].plot(instaAsymtab.median().values, color)
        tabAx[3, classed].plot(tailAnglestab.median().values, color)
      
      instaTBFmedian  = instaTBFtab.median().values
      instaAmpmedian  = instaAmptab.median().values
      instaAsymmedian = instaAsymtab.median().values
      
      dist = abs(instaTBFtab-instaTBFmedian).sum(axis=1)/abs(instaTBFmedian).sum() + abs(instaAmptab-instaAmpmedian).sum(axis=1)/abs(instaAmpmedian).sum() + abs(instaAsymtab-instaAsymmedian).sum(axis=1)/abs(instaAsymmedian).sum()
      
      # if modelUsedForClustering == 'GaussianMixture':
        # dist = 1 - dfTemp['classProba' + str(classed)]
        
      if len(dist):
        idMinDist = dist.infer_objects().idxmin()
      else:
        idMinDist = -1
      mostRepresentativeBout[idxCond, classed] = idMinDist
  
  if scaleGraphs:
    for classed in range(0, len(proportions[0])):
      if nbCluster == 1:
        tabAx[0].scatter(xaxis, freqAxis, None, 'w')
        tabAx[1].scatter(xaxis, ampAxis, None, 'w')
        tabAx[2].scatter(xaxis, asymAxis, None, 'w')
        tabAx[3].scatter(xaxis, angAxis, None, 'w')    
      else:
        tabAx[0, classed].scatter(xaxis, freqAxis, None, 'w')
        tabAx[1, classed].scatter(xaxis, ampAxis, None, 'w')
        tabAx[2, classed].scatter(xaxis, asymAxis, None, 'w')
        tabAx[3, classed].scatter(xaxis, angAxis, None, 'w')
  if nbCluster == 1:
    tabAx[0].legend()
    tabAx[0].set_ylabel('Avg Insta Frequency')
    tabAx[1].set_ylabel('Avg Insta Amplitude')
    tabAx[2].set_ylabel('Avg Insta Asymetry')
    tabAx[3].set_ylabel('Avg Angle')  
  else:
    tabAx[0, 0].legend()
    tabAx[0, 0].set_ylabel('Avg Insta Frequency')
    tabAx[1, 0].set_ylabel('Avg Insta Amplitude')
    tabAx[2, 0].set_ylabel('Avg Insta Asymetry')
    tabAx[3, 0].set_ylabel('Avg Angle')
  for i in range(0, nbCluster):
    labelX = "Cluster " + str(i+1) + "\n"
    for j, condName in enumerate(conditions):
      labelX = labelX + str(condName) + ": " + str(round(proportions[j,i]*100*100)/100) + "% (in " + possibleColorsNames[j] + ")\n"
    if nbCluster == 1:
      tabAx[3].set_xlabel(labelX)
    else:
      tabAx[3, i].set_xlabel(labelX)
  
  plt.savefig(os.path.join(outputFolderResult, 'medianValuesUsedForClusteringForEachClusterAndCondition.png'))
  
  if showFigures:
    plt.show()
  
  # Plot most representative bout for each cluster
  fig, tabAx2 = plt.subplots(4, len(proportions[0]), figsize=(22.9, 8.8))
  for cond in range(0, len(proportions)):
    for classed in range(0, len(proportions[0])):
      idMinDist = mostRepresentativeBout[cond, classed]
      if idMinDist != -1 and not(np.isnan(idMinDist)):
        instaTBFtab   = dfParam.loc[idMinDist, instaTBF]
        instaAmptab   = dfParam.loc[idMinDist, instaAmp]
        instaAsymtab  = dfParam.loc[idMinDist, instaAsym]
        tailAnglestab = dfParam.loc[idMinDist, tailAngles]
        color = possibleColors[cond]
        if nbCluster == 1:
          tabAx2[0].plot(instaTBFtab.values,  color)
          tabAx2[1].plot(instaAmptab.values,  color)
          tabAx2[2].plot(instaAsymtab.values, color)
          tabAx2[3].plot(tailAnglestab.values, color)
        else:
          tabAx2[0, classed].plot(instaTBFtab.values,  color)
          tabAx2[1, classed].plot(instaAmptab.values,  color)
          tabAx2[2, classed].plot(instaAsymtab.values, color)
          tabAx2[3, classed].plot(tailAnglestab.values, color)
  if scaleGraphs:
    for classed in range(0, len(proportions[0])):
      if nbCluster == 1:
        tabAx2[0].scatter(xaxis, freqAxis, None, 'w')
        tabAx2[1].scatter(xaxis, ampAxis, None, 'w')
        tabAx2[2].scatter(xaxis, asymAxis, None, 'w')
        tabAx2[3].scatter(xaxis, angAxis, None, 'w')      
      else:
        tabAx2[0, classed].scatter(xaxis, freqAxis, None, 'w')
        tabAx2[1, classed].scatter(xaxis, ampAxis, None, 'w')
        tabAx2[2, classed].scatter(xaxis, asymAxis, None, 'w')
        tabAx2[3, classed].scatter(xaxis, angAxis, None, 'w')
  if nbCluster == 1:
    tabAx2[0].set_ylabel('Avg Insta Frequency')
    tabAx2[1].set_ylabel('Avg Insta Amplitude')
    tabAx2[2].set_ylabel('Avg Insta Asymetry')
    tabAx2[3].set_ylabel('Avg Angle')  
  else:
    tabAx2[0, 0].set_ylabel('Avg Insta Frequency')
    tabAx2[1, 0].set_ylabel('Avg Insta Amplitude')
    tabAx2[2, 0].set_ylabel('Avg Insta Asymetry')
    tabAx2[3, 0].set_ylabel('Avg Angle')
  for i in range(0, nbCluster):
    labelX = "Most representative bout of cluster "+ str(i+1) + ":\n"
    for j, condName in enumerate(conditions):
      labelX = labelX + "for " + str(condName) + " (in " + possibleColorsNames[j] + ")\n"
    if nbCluster == 1:
      tabAx2[3].set_xlabel(labelX)
    else:
      tabAx2[3, i].set_xlabel(labelX)
  plt.savefig(os.path.join(outputFolderResult, 'mostRepresentativeBoutForEachClusterAndCondition.png'))
  if showFigures:
    plt.show()

  # Getting most representative sorted bouts
  sortedRepresentativeBouts = []
  sortedRepresentativeBoutsSpeed = []
  sortedRepresentativeBoutsHeadingDiff = []
  sortedRepresentativeBoutsIndex = []
  for classed in range(0, len(proportions[0])):
    dfTemp = dfParam.loc[(dfParam['classification'] == classed)]
    instaTBFtab   = dfTemp[instaTBF]
    instaAmptab   = dfTemp[instaAmp]
    instaAsymtab  = dfTemp[instaAsym]
    tailAnglestab = dfTemp[tailAngles]
    
    instaTBFmedian  = instaTBFtab.median().values
    instaAmpmedian  = instaAmptab.median().values
    instaAsymmedian = instaAsymtab.median().values

    dist = abs(instaTBFtab-instaTBFmedian).sum(axis=1)/abs(instaTBFmedian).sum() + abs(instaAmptab-instaAmpmedian).sum(axis=1)/abs(instaAmpmedian).sum() + abs(instaAsymtab-instaAsymmedian).sum(axis=1)/abs(instaAsymmedian).sum()
    # if modelUsedForClustering == 'GaussianMixture':
      # dist = 1 - dfTemp['classProba' + str(classed)]
    
    if tailAngles[0] in dfParam.columns:
      sortedRepresentativeBouts.append(dfParam.loc[dist.index.values[dist.values.argsort()], tailAngles])
    if instaSpeed[0] in dfParam.columns:
      sortedRepresentativeBoutsSpeed.append(dfParam.loc[dist.index.values[dist.values.argsort()], instaSpeed])
    if instaHeadingDiff[0] in dfParam.columns:
      sortedRepresentativeBoutsHeadingDiff.append(dfParam.loc[dist.index.values[dist.values.argsort()], instaHeadingDiff])
    sortedRepresentativeBoutsIndex.append(dist.index.values[dist.values.argsort()])
  
  # Plot most representative bouts
  nbOfMostRepresentativeBoutsToPlot = 10000000000000
  for classed in range(0, len(proportions[0])):
    nb = len(sortedRepresentativeBouts[classed].index)
    if nb < nbOfMostRepresentativeBoutsToPlot:
      nbOfMostRepresentativeBoutsToPlot = nb
  if nbOfMostRepresentativeBoutsToPlot > 100:
    nbOfMostRepresentativeBoutsToPlot = 100
  fig, tabAx3 = plt.subplots(len(proportions[0]),1, figsize=(22.9, 8.8))
  for classed in range(0, len(proportions[0])):
    indices = sortedRepresentativeBouts[classed].index
    for j in range(0, nbOfMostRepresentativeBoutsToPlot):
      tailAnglestab = sortedRepresentativeBouts[classed].loc[indices[j]].values
      color = 'b'
      if nbCluster == 1:
        tabAx3.plot(tailAnglestab, color)
      else:
        tabAx3[classed].plot(tailAnglestab, color)
  for i in range(0,len(proportions[0])):
    if nbCluster == 1:
      tabAx3.set_ylabel('Cluster '+str(i+1))
    else:
      tabAx3[i].set_ylabel('Cluster '+str(i+1))
  if nbCluster == 1:
    tabAx3.set_xlabel("Tail angle over time for the\n"+str(nbOfMostRepresentativeBoutsToPlot)+' most representative bouts for each cluster')
  else:
    tabAx3[len(proportions[0])-1].set_xlabel("Tail angle over time for the\n"+str(nbOfMostRepresentativeBoutsToPlot)+' most representative bouts for each cluster')
  plt.savefig(os.path.join(outputFolderResult, str(nbOfMostRepresentativeBoutsToPlot) + 'mostRepresentativeBoutsForEachCluster.png'))
  if showFigures:
    plt.show()
  
  # Plot most representative bouts - second plot
  fig, tabAx3 = plt.subplots(len(proportions[0]),1, figsize=(22.9, 8.8))
  for classed in range(0, len(proportions[0])):
    nbOfMostRepresentativeBoutsToPlot = 10000000000000
    nbOfMostRepresentativeBoutsToPlot = len(sortedRepresentativeBouts[classed].index)
    if nbOfMostRepresentativeBoutsToPlot > 100:
      nbOfMostRepresentativeBoutsToPlot = 100
    indices = sortedRepresentativeBouts[classed].index
    for j in range(0, nbOfMostRepresentativeBoutsToPlot):
      tailAnglestab = sortedRepresentativeBouts[classed].loc[indices[j]].values
      color = 'b'
      if nbCluster == 1:
        tabAx3.plot(tailAnglestab, color)
      else:
        tabAx3[classed].plot(tailAnglestab, color)
      
  for i in range(0,len(proportions[0])):
    if nbCluster == 1:
      tabAx3.set_ylabel('Cluster '+str(i+1))
    else:
      tabAx3[i].set_ylabel('Cluster '+str(i+1))
  if nbCluster == 1:
    tabAx3.set_xlabel("Tail angle over time for the most representative bouts for each cluster")
  else:
    tabAx3[len(proportions[0])-1].set_xlabel("Tail angle over time for the most representative bouts for each cluster")
  plt.savefig(os.path.join(outputFolderResult, 'mostRepresentativeBoutsForEachCluster.png'))
  if showFigures:
    plt.show()
  
  
  # Plot most representative bouts: Speed
  if len(sortedRepresentativeBoutsSpeed):
    nbOfMostRepresentativeBoutsToPlot = 10000000000000
    for classed in range(0, len(proportions[0])):
      nb = len(sortedRepresentativeBoutsSpeed[classed].index)
      if nb < nbOfMostRepresentativeBoutsToPlot:
        nbOfMostRepresentativeBoutsToPlot = nb
    if nbOfMostRepresentativeBoutsToPlot > 10:
      nbOfMostRepresentativeBoutsToPlot = 10
    fig, tabAx3 = plt.subplots(len(proportions[0]),1, figsize=(22.9, 8.8))
    for classed in range(0, len(proportions[0])):
      indices = sortedRepresentativeBoutsSpeed[classed].index
      for j in range(0, nbOfMostRepresentativeBoutsToPlot):
        tailAnglestab = sortedRepresentativeBoutsSpeed[classed].loc[indices[j]].values
        color = 'b'
        if nbCluster == 1:
          tabAx3.plot(tailAnglestab, color)
        else:
          tabAx3[classed].plot(tailAnglestab, color)
    for i in range(0,len(proportions[0])):
      if nbCluster == 1:
        tabAx3.set_ylabel('Cluster '+str(i+1))
      else:
        tabAx3[i].set_ylabel('Cluster '+str(i+1))
    if nbCluster == 1:
      tabAx3.set_xlabel("Tail angle over time for the\n"+str(nbOfMostRepresentativeBoutsToPlot)+' most representative bouts for each cluster')
    else:
      tabAx3[len(proportions[0])-1].set_xlabel("Tail angle over time for the\n"+str(nbOfMostRepresentativeBoutsToPlot)+' most representative bouts for each cluster')
    plt.savefig(os.path.join(outputFolderResult, str(nbOfMostRepresentativeBoutsToPlot) + 'mostRepresentativeBoutsForEachClusterSpeed.png'))
    if showFigures:
      plt.show()
  

  # Plot most representative bouts: Heading diff
  if len(sortedRepresentativeBoutsHeadingDiff):
    nbOfMostRepresentativeBoutsToPlot = 10000000000000
    for classed in range(0, len(proportions[0])):
      nb = len(sortedRepresentativeBoutsHeadingDiff[classed].index)
      if nb < nbOfMostRepresentativeBoutsToPlot:
        nbOfMostRepresentativeBoutsToPlot = nb
    if nbOfMostRepresentativeBoutsToPlot > 10:
      nbOfMostRepresentativeBoutsToPlot = 10
    fig, tabAx3 = plt.subplots(len(proportions[0]),1, figsize=(22.9, 8.8))
    for classed in range(0, len(proportions[0])):
      indices = sortedRepresentativeBoutsHeadingDiff[classed].index
      for j in range(0, nbOfMostRepresentativeBoutsToPlot):
        tailAnglestab = sortedRepresentativeBoutsHeadingDiff[classed].loc[indices[j]].values
        color = 'b'
        if nbCluster == 1:
          tabAx3.plot(tailAnglestab, color)
        else:
          tabAx3[classed].plot(tailAnglestab, color)
    for i in range(0,len(proportions[0])):
      if nbCluster == 1:
        tabAx3.set_ylabel('Cluster '+str(i+1))
      else:
        tabAx3[i].set_ylabel('Cluster '+str(i+1))
    if nbCluster == 1:
      tabAx3.set_xlabel("Tail angle over time for the\n"+str(nbOfMostRepresentativeBoutsToPlot)+' most representative bouts for each cluster')
    else:
      tabAx3[len(proportions[0])-1].set_xlabel("Tail angle over time for the\n"+str(nbOfMostRepresentativeBoutsToPlot)+' most representative bouts for each cluster')
    plt.savefig(os.path.join(outputFolderResult, str(nbOfMostRepresentativeBoutsToPlot) + 'mostRepresentativeBoutsForEachClusterHeadingDiff.png'))
    if showFigures:
      plt.show()
  
  
  # Creating validation videos: Beginning (10 movements each)
  if videoSaveFirstTenBouts:
    length = 150
    for boutCategory in range(0, nbCluster):
      print("boutCategory:",boutCategory+1)
      out = cv2.VideoWriter(os.path.join(outputFolderResult, 'cluster' + str(boutCategory+1) + '.avi'),cv2.VideoWriter_fourcc('M','J','P','G'), 10, (length,length))
      indices = sortedRepresentativeBouts[boutCategory].index
      nbTemp = len(indices)
      if nbTemp < nbVideosToSave:
        nbVideosToSave = nbTemp
      
      r = [i for i in range(0, nbVideosToSave)]
      for num in r:
        print("num:",num)
        BoutStart = int(dfParam.loc[indices[num],'BoutStart'])
        BoutEnd   = int(dfParam.loc[indices[num],'BoutEnd'])
        Well_ID   = int(dfParam.loc[indices[num],'Well_ID'])
        Trial_ID  = dfParam.loc[indices[num],'Trial_ID']
        NumBout   = dfParam.loc[indices[num],'NumBout']
        out = outputValidationVideo(pathToVideos, Trial_ID, '.txt', Well_ID, NumBout, 1, BoutStart, BoutEnd, out, length, analyzeAllWellsAtTheSameTime, ZZoutputLocation)
      
      out.release()
  
  # Looking into global parameters
  if globalParametersCalculations:
    
    for plotOutliersAndMean in [0, 1]:
      globParam1 = ['Bout Duration (s)', 'Bout Distance (mm)', 'Bout Speed (mm/s)', 'Number of Oscillations', 'meanTBF', 'maxTailAngleAmplitude']
      globParam2 = ['Max TBF (Hz)', 'Mean TBF (Hz)', 'medianOfInstantaneousTBF', 'Max absolute TBA (deg.)', 'Mean absolute TBA (deg.)', 'Median absolute TBA (deg.)']
      globParam3 = ['Absolute Yaw (deg)', 'TBA#1 timing (s)', 'TBA#1 Amplitude (deg)', 'IBI (s)', 'xmean', 'ymean']
      globParam4 = ['binaryClass25degMaxTailAngle', 'BoutFrameNumberStart', 'tailAngleSymmetry', 'secondBendAmpDividedByFirst']
      
      for idxGlobParam, globParam in enumerate([globParam1, globParam2, globParam3, globParam4]):
        fig, tabAx = plt.subplots(2, 3, figsize=(22.9, 8.8))
        for idx, parameter in enumerate(globParam):
          concatenatedValues = []
          for boutCategory in range(0, nbCluster):
            indices = sortedRepresentativeBouts[boutCategory].index
            values  = dfParam.loc[indices[:],parameter].values
            concatenatedValues.append(values)
          tabAx[int(idx/3), idx%3].set_title(parameter)
          tabAx[int(idx/3), idx%3].boxplot(concatenatedValues, showmeans=1, showfliers=plotOutliersAndMean)
        globParamFileName = 'globalParametersforEachCluster' + str(idxGlobParam) + '.png' if plotOutliersAndMean else 'globalParametersforEachCluster_NoOutliers_' + str(idxGlobParam) + '.png'
        plt.savefig(os.path.join(outputFolderResult, globParamFileName))
        if showFigures:
          plt.plot()
          plt.show()
    
    ###
    
    globParam1 = ['Bout Duration (s)', 'Bout Distance (mm)', 'Bout Speed (mm/s)', 'Number of Oscillations', 'meanTBF', 'maxTailAngleAmplitude']
    globParam2 = ['Max TBF (Hz)', 'Mean TBF (Hz)', 'medianOfInstantaneousTBF', 'Max absolute TBA (deg.)', 'Mean absolute TBA (deg.)', 'Median absolute TBA (deg.)']
    globParam3 = ['Absolute Yaw (deg)', 'TBA#1 timing (s)', 'TBA#1 Amplitude (deg)', 'IBI (s)', 'xmean', 'ymean']
    globParam4 = ['binaryClass25degMaxTailAngle', 'BoutFrameNumberStart', 'tailAngleSymmetry', 'secondBendAmpDividedByFirst']
    
    for calculateKinematicParametersPerFish in [True, False]:
    
      if calculateKinematicParametersPerFish:
        globParamTot = globParam1 + globParam2 + globParam3 + globParam4
        dfKinematicInsideCluster = dfParam[['Trial_ID', 'Well_ID', 'classification', 'Condition'] + globParamTot].astype({col: 'float' for col in globParamTot}).groupby(['Trial_ID', 'Well_ID', 'classification', 'Condition']).median()
        dfKinematicInsideCluster['classification'] = [elem[2] for elem in dfKinematicInsideCluster.index]
        dfKinematicInsideCluster['Condition']      = [elem[3] for elem in dfKinematicInsideCluster.index]
      else:
        dfKinematicInsideCluster = dfParam
      
      conditions = dfParam["Condition"].unique().tolist()
      palette = dict(zip(sortGenotypes(conditions), sns.color_palette(n_colors=len(conditions))))
      for idxGlobParam, globParam in enumerate([globParam1, globParam2, globParam3, globParam4]):
        fig, tabAx = plt.subplots(2, 3, figsize=(22.9, 8.8))
        fig.tight_layout(pad=3.0)
        for idx, parameter in enumerate(globParam):
          b = sns.boxplot(ax=tabAx[int(idx/3), idx%3], data=dfKinematicInsideCluster, x="classification", y=parameter, hue="Condition", showfliers = False, hue_order=palette.keys(), palette=palette)
          b.set_ylabel('', fontsize=0)
          b.set_xlabel('', fontsize=0)
          b.axes.set_title(parameter,fontsize=30)
        globParamFileName = 'globalParametersforEachCluster_Conditions_NoOutliers_' + str(idxGlobParam)
        if calculateKinematicParametersPerFish:
          globParamFileName = globParamFileName + '_PerFish.png'
        else:
          globParamFileName = globParamFileName + '_PerBout.png'
        plt.savefig(os.path.join(outputFolderResult, globParamFileName))
        if showFigures:
          plt.plot()
          plt.show()
  
  
  # Saves classifications
  if 'classProba0' in dfParam.columns.tolist():
    dfParam[['Trial_ID','Well_ID','NumBout','Condition', 'Genotype', 'classification'] + ['classProba' + str(j) for j in range(0, nbCluster)]].to_csv(os.path.join(outputFolderResult, 'classifications.txt'))
  else:
    dfParam[['Trial_ID','Well_ID','NumBout','Condition', 'Genotype', 'classification']].to_csv(os.path.join(outputFolderResult, 'classifications.txt'))
  
  
  # Plotting mostToLeastRepresentativeBoutsForEachCluster figure
  
  optimalLength = 12
  
  fig, tabAx3 = plt.subplots(3, len(proportions[0]), figsize=(22.9, 8.8))
  for classed in range(0, len(proportions[0])):
    indices = sortedRepresentativeBouts[classed].index
    for j in range(0, 3):
      ind2size = optimalLength if len(indices) >= 3*optimalLength else max(1, int(len(indices) / 3))
      if j == 0:
        indices2 = indices[0:ind2size]
      elif j == 1:
        indices2 = indices[int(len(indices)/2) - int(ind2size/2) : int(len(indices)/2) + int(ind2size/2)]
      else:
        indices2 = indices[len(indices) - ind2size : len(indices)]
      for k in range(0, len(indices2)):
        tailAnglestab = sortedRepresentativeBouts[classed].loc[indices2[k]].values
        color = 'b'
        if nbCluster == 1:
          tabAx3[j].plot(tailAnglestab, color)
        else:
          tabAx3[j, classed].plot(tailAnglestab, color)
  for i in range(0, len(proportions[0])):
    if nbCluster == 1:
      tabAx3[2].set_xlabel("Bouts of cluster 1")
    else:
      tabAx3[2, i].set_xlabel("Bouts of cluster " + str(i + 1))
  if nbCluster == 1:
    tabAx3[0].set_ylabel("Most representative bouts")
    tabAx3[1].set_ylabel("In between")
    tabAx3[2].set_ylabel("Least representative bouts")
  else:
    tabAx3[0, 0].set_ylabel("Most representative bouts")
    tabAx3[1, 0].set_ylabel("In between")
    tabAx3[2, 0].set_ylabel("Least representative bouts")
  plt.savefig(os.path.join(outputFolderResult, 'mostToLeastRepresentativeBoutsForEachCluster.png'))
  
  # Plotting bouts one by one for each cluster and writing bout indices in txt file
  if False:
    outF = open(os.path.join(outputFolderResult, 'boutIndices.txt'), "w")
    for classed in range(0, len(proportions[0])):
      indices = sortedRepresentativeBouts[classed].index
      for j in range(0, 3):
        ind2size = optimalLength if len(indices) >= 3*optimalLength else max(1, int(len(indices) / 3))
        if j == 0:
          indices2 = indices[0:ind2size]
        elif j == 1:
          indices2 = indices[int(len(indices)/2) - int(ind2size/2) : int(len(indices)/2) + int(ind2size/2)]
        else:
          indices2 = indices[len(indices) - ind2size : len(indices)]
        # Plot
        fig, tabAx3 = plt.subplots(4, 3, figsize=(22.9, 8.8))
        for k in range(0, len(indices2)):
          tailAnglestab = dfParam.loc[indices2[k], tailAngles].tolist()
          tabAx3[int(k / 3), k % 3].plot(tailAnglestab, 'b', label = str(indices2[k]))
          tabAx3[int(k / 3), k % 3].legend()
          tabAx3[int(k / 3), k % 3].title.set_text(str(indices2[k]))
        region = 'mostRepresentative' if j == 0 else 'inBetween' if j == 1 else 'leastRepresentative'
        plt.savefig(os.path.join(outputFolderResult, 'cluster' + str(classed + 1) + '_' + region + '.png'))

  return [proportions, sortedRepresentativeBouts, sortedRepresentativeBoutsIndex]


from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.utilities import calculateAngle, distBetweenThetas
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.detectMovementWithTrackedDataAfterTracking import detectMovementWithTrackedDataAfterTracking
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.detectMovementWithRawVideoInsideTracking import detectMovementWithRawVideoInsideTracking
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.getListOfWellsOnWhichToRunTheTracking import getListOfWellsOnWhichToRunTheTracking
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.backgroundSubtractionOnWholeImage import backgroundSubtractionOnWholeImage
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.backgroundSubtractionOnlyOnROIs import backgroundSubtractionOnlyOnROIs
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
from zebrazoom.code.extractParameters import extractParameters
import zebrazoom.code.util as util
import zebrazoom.code.tracking
import numpy as np
import queue
import math
import time
import cv2

class Tracking(zebrazoom.code.tracking.BaseTrackingMethod):
  
  def __init__(self, videoPath, wellPositions, hyperparameters):
    self._videoPath = videoPath
    self._wellPositions = wellPositions
    self._hyperparameters = hyperparameters
    self._auDessusPerAnimalIdList = None
    self._firstFrame = self._hyperparameters["firstFrame"]
    self._lastFrame = self._hyperparameters["lastFrame"]
    self._nbTailPoints = self._hyperparameters["nbTailPoints"]
    self._previousFrames = None
    self._trackingDataPerWell = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1, self._nbTailPoints, 2)) for _ in range(len(self._wellPositions))]
    self._lastFirstTheta = np.zeros(len(self._wellPositions))
    self._lastFirstTheta[:] = -99999
    self._listOfWellsOnWhichToRunTheTracking = [i for i in range(0, len(self._wellPositions))] if hyperparameters["onlyTrackThisOneWell"] == -1 else [hyperparameters["onlyTrackThisOneWell"]]
    self._times2 = np.zeros((self._lastFrame - self._firstFrame + 1, 5))
    self._printInterTime = False


  def run(self):
    
    ### Step 1 (out of 2): Tracking:
    resizeFrameFactor = self._hyperparameters["resizeFrameFactor"] if "resizeFrameFactor" in self._hyperparameters else 0
    
    # Getting video reader
    cap = zzVideoReading.VideoCapture(self._videoPath)
    if (cap.isOpened()== False):
      print("Error opening video stream or file")
    
    # Simple background extraction with first and last frame of the video + Getting list of wells on which to run the tracking
    ret, self._background = cap.read()
    if resizeFrameFactor:
      self._background = cv2.resize(self._background, (int(len(self._background[0])/resizeFrameFactor), int(len(self._background)/resizeFrameFactor)))
    cap.set(cv2.CAP_PROP_POS_FRAMES, int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) - 1)
    ret, frame = cap.read()
    if resizeFrameFactor:
      frame = cv2.resize(frame, (int(len(frame[0])/resizeFrameFactor), int(len(frame)/resizeFrameFactor)))
    if self._hyperparameters["chooseWellsToRunTrackingOnWithFirstAndLastFrame"]:
      self._listOfWellsOnWhichToRunTheTracking = getListOfWellsOnWhichToRunTheTracking(self, self._background[:,:,0], frame[:,:,0])
    print("listOfWellsOnWhichToRunTheTracking:", self._listOfWellsOnWhichToRunTheTracking)
    self._background = cv2.max(frame, self._background) # INCONSISTENT!!! should be changed!
    self._background = cv2.cvtColor(self._background, cv2.COLOR_BGR2GRAY) # INCONSISTENT!!! should be changed!
    cap.set(cv2.CAP_PROP_POS_FRAMES, self._firstFrame)
    
    if resizeFrameFactor:
      initialWellPositions = self._wellPositions.copy()
      self._wellPositions = [{'topLeftX': int(pos['topLeftX']/resizeFrameFactor), 'topLeftY': int(pos['topLeftY']/resizeFrameFactor), 'lengthX': int(pos['lengthX']/resizeFrameFactor), 'lengthY': int(pos['lengthY']/resizeFrameFactor)} for pos in self._wellPositions]
    
    # Initializing variables
    times  = np.zeros((self._lastFrame - self._firstFrame + 1, 2))
    ret = True
    
    # Going through each frame of the video
    startTime = time.time()
    k = self._firstFrame
    while (ret and k <= self._lastFrame):
      time1 = time.time()
      ret, frame = cap.read()
      time2 = time.time()
      if resizeFrameFactor:
        frame = cv2.resize(frame, (int(len(frame[0])/resizeFrameFactor), int(len(frame)/resizeFrameFactor)))
      if ret:
        if self._hyperparameters["backgroundSubtractionOnWholeImage"] or k == self._firstFrame:
          backgroundSubtractionOnWholeImage(self, frame, k-self._firstFrame)
        else:
          backgroundSubtractionOnlyOnROIs(self, frame, k-self._firstFrame)
      
      time3 = time.time()
      times[k-self._firstFrame, 0] = time2 - time1
      times[k-self._firstFrame, 1] = time3 - time2
      k += 1
    
    if resizeFrameFactor:
      self._trackingDataPerWell = [resizeFrameFactor * elem for elem in self._trackingDataPerWell]
      self._wellPositions = initialWellPositions
    
    endTime = time.time()
    
    cap.release()
    
    print("")
    print("Color to grey:"           , np.median(self._times2[:,0]))
    print("Bout detection:"          , np.median(self._times2[:,1]))
    print("Background substraction:" , np.median(self._times2[:,2]))
    print("Gaussian blur:"           , np.median(self._times2[:,3]))
    print("Tracking on each well:"   , np.median(self._times2[:,4]))
    
    loadingImagesTime       = np.median(times[:,0])
    processingImagesTime    = np.median(times[:,1])
    percentTimeSpentLoading = loadingImagesTime / (loadingImagesTime + processingImagesTime)
    print("Median time spent on: Loading images:", loadingImagesTime, "; Processing images:", processingImagesTime)
    print("Percentage of time spent loading images:", percentTimeSpentLoading*100)
    print("Total tracking Time:", endTime - startTime)
    print("Tracking Time (without loading image):", (endTime - startTime) * (1 - percentTimeSpentLoading))
    print("Total tracking fps:", k / (endTime - startTime))
    print("Tracking fps (without loading image):", k / ((endTime - startTime) * (1 - percentTimeSpentLoading)))
    print("")
    
    ### Step 2 (out of 2): Extracting bout of movements:
    
    if self._hyperparameters["detectMovementWithRawVideoInsideTracking"] and self._hyperparameters["thresForDetectMovementWithRawVideo"]:
    
      trackingHeadingAllAnimalsList = [[[((calculateAngle(self._trackingDataPerWell[wellNumber][animalNumber][i][0][0], self._trackingDataPerWell[wellNumber][animalNumber][i][0][1], self._trackingDataPerWell[wellNumber][animalNumber][i][1][0], self._trackingDataPerWell[wellNumber][animalNumber][i][1][1]) + math.pi) % (2 * math.pi) if len(self._trackingDataPerWell[wellNumber][0][i]) > 1 else 0) for i in range(0, self._lastFrame-self._firstFrame+1)] for animalNumber in range(0, self._hyperparameters["nbAnimalsPerWell"])] for wellNumber in range(0, len(self._wellPositions))]
      
      return {wellNumber: extractParameters([self._trackingDataPerWell[wellNumber], trackingHeadingAllAnimalsList[wellNumber], [], 0, 0, self._auDessusPerAnimalIdList[wellNumber]], wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background) for wellNumber in range(0, len(self._wellPositions))}

    elif self._hyperparameters["adjustDetectMovWithRawVideo"]:

      trackingHeadingAllAnimalsList = [[[((calculateAngle(self._trackingDataPerWell[wellNumber][animalNumber][i][0][0], self._trackingDataPerWell[wellNumber][animalNumber][i][0][1], self._trackingDataPerWell[wellNumber][animalNumber][i][1][0], self._trackingDataPerWell[wellNumber][animalNumber][i][1][1]) + math.pi) % (2 * math.pi) if len(self._trackingDataPerWell[wellNumber][0][i]) > 1 else 0) for i in range(0, self._lastFrame-self._firstFrame+1)] for animalNumber in range(0, self._hyperparameters["nbAnimalsPerWell"])] for wellNumber in range(0, len(self._wellPositions))]

      return {wellNumber: extractParameters([self._trackingDataPerWell[wellNumber], trackingHeadingAllAnimalsList[wellNumber], [], 0, 0], wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background) for wellNumber in self._listOfWellsOnWhichToRunTheTracking}

    elif  self._hyperparameters["coordinatesOnlyBoutDetection"]:

      trackingHeadingAllAnimalsList = [[[((calculateAngle(self._trackingDataPerWell[wellNumber][animalNumber][i][0][0], self._trackingDataPerWell[wellNumber][animalNumber][i][0][1], self._trackingDataPerWell[wellNumber][animalNumber][i][1][0], self._trackingDataPerWell[wellNumber][animalNumber][i][1][1]) + math.pi) % (2 * math.pi) if len(self._trackingDataPerWell[wellNumber][0][i]) > 1 else 0) for i in range(0, self._lastFrame-self._firstFrame+1)] for animalNumber in range(0, self._hyperparameters["nbAnimalsPerWell"])] for wellNumber in range(0, len(self._wellPositions))]

      return {wellNumber: extractParameters([self._trackingDataPerWell[wellNumber], trackingHeadingAllAnimalsList[wellNumber], [], 0, 0], wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background) for wellNumber in range(0, len(self._wellPositions))}

    else:
    
      if not(self._hyperparameters["detectBouts"]):
        
        trackingHeadingAllAnimalsList = [[[((calculateAngle(self._trackingDataPerWell[wellNumber][animalNumber][i][0][0], self._trackingDataPerWell[wellNumber][animalNumber][i][0][1], self._trackingDataPerWell[wellNumber][animalNumber][i][1][0], self._trackingDataPerWell[wellNumber][animalNumber][i][1][1]) + math.pi) % (2 * math.pi) if len(self._trackingDataPerWell[wellNumber][0][i]) > 1 else 0) for i in range(0, self._lastFrame-self._firstFrame+1)] for animalNumber in range(0, self._hyperparameters["nbAnimalsPerWell"])] for wellNumber in range(0, len(self._wellPositions))]
        
        self._auDessusPerAnimalIdList = [[np.ones((self._lastFrame-self._firstFrame+1, 1)) for nbAnimalsPerWell in range(0, self._hyperparameters["nbAnimalsPerWell"])] for wellNumber in range(len(self._wellPositions))]
        
        return {wellNumber: extractParameters([self._trackingDataPerWell[wellNumber], trackingHeadingAllAnimalsList[wellNumber], [], 0, 0, self._auDessusPerAnimalIdList[wellNumber]], wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background) for wellNumber in range(0, len(self._wellPositions))}
        
      else:
        
        outputData = detectMovementWithTrackedDataAfterTracking(self)
      
      return outputData


zebrazoom.code.tracking.register_tracking_method('fastFishTracking.tracking', Tracking)



import json
import matplotlib.pyplot as plt
import cv2
import numpy as np
import math
import pandas as pd
import os

from PyQt5.QtWidgets import QFileDialog


def createDistanceBetweenFramesExcelFile(ZZoutputFolder, fps, pixelSize):
  
  pathToResultFolder =  QFileDialog.getExistingDirectory(caption='Choose the result folder for which an excel file should be generated', directory=ZZoutputFolder)
  if not pathToResultFolder:
    return

  videoName = os.path.basename(os.path.normpath(pathToResultFolder))
  
  numberOfWells       = 0
  totalNumberOfFrames = 0
  
  # Transfering data from results_videoName.txt to 'per frame' pandas dataframe
  with open(os.path.join(pathToResultFolder, 'results_' + videoName + '.txt')) as video:
  
    supstruct = json.load(video)
    
    numberOfWells = len(supstruct['wellPoissMouv'])
    firstFrame = 0
    lastFrame  = 0
    if ('lastFrame' in supstruct) and ('firstFrame' in supstruct):
      totalNumberOfFrames = supstruct['lastFrame'] - supstruct['firstFrame'] + 1
      firstFrame = int(supstruct['firstFrame'])
      lastFrame  = int(supstruct['lastFrame'])
    else:
      raise NameError('firstFrame and lastFrame are not in the superstructure results file!')
    
    print("totalNumberOfFrames: ", totalNumberOfFrames)

    data = pd.DataFrame(data=[[wellNum, frameNum, frameNum/fps, (frameNum/fps)/60, float('nan'), float('nan'), 0] for wellNum in range(0, numberOfWells) for frameNum in range(firstFrame, lastFrame)], columns=['wellId', 'frameNumber', 'second', 'minute', 'xPosition', 'yPosition', 'instantaneousDistance'])
    
    for numWell in range(0, len(supstruct['wellPoissMouv'])):  
      for numAnimal in range(0, len(supstruct['wellPoissMouv'][numWell])):
        for numBout in range(0, len(supstruct['wellPoissMouv'][numWell][numAnimal])):
          bout = supstruct['wellPoissMouv'][numWell][numAnimal][numBout]
          # if bout["BoutEnd"] < totalNumberOfFrames:
          print("numWell:", numWell, "; numBout:", numBout, "; boutStart:", bout["BoutStart"])
          try:
            data.loc[np.logical_and(np.logical_and(data['wellId'] == numWell, data['frameNumber'] >= bout["BoutStart"]), data['frameNumber'] <= bout["BoutEnd"]), 'xPosition'] = bout['HeadX']
            data.loc[np.logical_and(np.logical_and(data['wellId'] == numWell, data['frameNumber'] >= bout["BoutStart"]), data['frameNumber'] <= bout["BoutEnd"]), 'yPosition'] = bout['HeadY']
          except:
            data.loc[np.logical_and(np.logical_and(data['wellId'] == numWell, data['frameNumber'] >= bout["BoutStart"]), data['frameNumber'] <= bout["BoutEnd"]), 'xPosition'] = bout['HeadX'][0:len(data.loc[np.logical_and(np.logical_and(data['wellId'] == numWell, data['frameNumber'] >= bout["BoutStart"]), data['frameNumber'] <= bout["BoutEnd"]), 'xPosition'])]
            data.loc[np.logical_and(np.logical_and(data['wellId'] == numWell, data['frameNumber'] >= bout["BoutStart"]), data['frameNumber'] <= bout["BoutEnd"]), 'yPosition'] = bout['HeadY'][0:len(data.loc[np.logical_and(np.logical_and(data['wellId'] == numWell, data['frameNumber'] >= bout["BoutStart"]), data['frameNumber'] <= bout["BoutEnd"]), 'yPosition'])]
  
  # Calculating the "instantaneousDistance"
  previousNumWell = 0
  previousXPos = -1
  previousYPos = -1
  for i in range(0, len(data) - 1):
    print(i, len(data))
    if data.loc[i, 'wellId'] == previousNumWell:
      if not(np.isnan(data.loc[i, 'xPosition'])):
        if not(previousXPos == -1) and not(previousYPos == -1):
          data.loc[i, 'instantaneousDistance'] = math.sqrt((data.loc[i, 'xPosition'] - previousXPos)**2 + (data.loc[i, 'yPosition'] - previousYPos)**2) * pixelSize
        previousXPos = data.loc[i, 'xPosition']
        previousYPos = data.loc[i, 'yPosition']
    else:
      previousNumWell = data.loc[i, 'wellId']
      previousXPos = -1
      previousYPos = -1
  
  # Changing the shape of the data
  finalDataArray = 0
  for wellId in range(0, numberOfWells):
    wellData = data.loc[data['wellId'] == wellId]
    if type(finalDataArray) == int:
      finalDataArray = np.concatenate((np.transpose(np.array([[i for i in range(firstFrame, lastFrame)]])), np.transpose(np.array([[i/fps for i in range(firstFrame, lastFrame)]])), np.transpose(np.array([wellData['instantaneousDistance']])), np.transpose(np.array([wellData['xPosition']])), np.transpose(np.array([wellData['yPosition']]))), axis=1)
    else:
      finalDataArray = np.concatenate((finalDataArray, np.transpose(np.array([wellData['instantaneousDistance']])), np.transpose(np.array([wellData['xPosition']])), np.transpose(np.array([wellData['yPosition']]))), axis=1)
  
  # Converting to excel
  finalData = pd.DataFrame(data=finalDataArray, columns=['frame', 'second'] + [item for i in range(0, numberOfWells) for item in ['distanceWell' + str(i), 'xPosition' + str(i), 'yPosition' + str(i)]])
  finalData.to_excel(os.path.join(pathToResultFolder, videoName + '_Distances.xlsx'))




import matplotlib.pyplot as plt
import numpy as np


def plotTailAngleHeatmap(tailAngleHeatmap: list, startFrame: int, tailLength: float, videoFPS: float, videoPixelSize: float):
  fig = plt.figure(1)
  maxAngle = np.nanmax(np.abs(tailAngleHeatmap))
  plt.pcolor(tailAngleHeatmap[::-1], vmin=-maxAngle, vmax=maxAngle)
  ax, *_ = fig.axes
  ax.set_xlabel('Time (in seconds)')
  ax.set_ylabel('Tail angle: Tail base to tail extremity (in mm)')
  ax.xaxis.set_major_formatter(lambda x, pos: f'{(x + startFrame) / videoFPS:.2f}')
  yStepSize = tailLength / len(tailAngleHeatmap)
  ax.yaxis.set_major_formatter(lambda y, pos: f'{y * yStepSize * videoPixelSize:.2f}')
  plt.colorbar()
  plt.show()



from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.utilities import appendPoint, distBetweenThetas, assignValueIfBetweenRange, calculateAngle
from scipy.interpolate import interp1d
import zebrazoom.code.util as util
from numpy import linspace
import numpy as np
import math
import cv2

def __insideTrackTail(depth, headPosition, frame, points, angle, maxDepth, steps, nbList, hyperparameters, debug, lenX, lenY, secondTry=False):
  
  x = headPosition[0]
  y = headPosition[1]
  
  pixSurMax = hyperparameters["headEmbededParamTailDescentPixThreshStop"]
  
  pixTotList = []
  
  distSubsquentPoints = 0.000001
  pixSur = 0
  
  while (distSubsquentPoints > 0 and depth < maxDepth and ((pixSur < pixSurMax) or (depth < hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth))):
  
    if depth == 0 and angle == -99999: # First frame, first segment
      thetaDiffAccept = 3.6
    else:
      if (depth < hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth):
        thetaDiffAccept = hyperparameters["thetaDiffAccept"]
      else:
        if not("authorizedRelativeLengthTailEnd2" in hyperparameters) or (depth < hyperparameters["authorizedRelativeLengthTailEnd2"]*maxDepth):
          thetaDiffAccept = hyperparameters["thetaDiffAcceptAfterAuthorizedRelativeLengthTailEnd"]
          nbList = hyperparameters["nbListAfterAuthorizedRelativeLengthTailEnd"]
        else:
          thetaDiffAccept = hyperparameters["thetaDiffAcceptAfterAuthorizedRelativeLengthTailEnd2"]
          nbList = hyperparameters["nbListAfterAuthorizedRelativeLengthTailEnd2"]
          

    pixTotMax = 1000000
    maxTheta  = angle

    l = [i*(math.pi/nbList) for i in range(0,2*nbList) if distBetweenThetas(i*(math.pi/nbList), angle) < thetaDiffAccept]

    for step in steps:
      if (step < maxDepth - depth) or (step == steps[0]):
        for theta in l:
          xNew = assignValueIfBetweenRange(int(x + step * (math.cos(theta))), 0, lenX)
          yNew = assignValueIfBetweenRange(int(y + step * (math.sin(theta))), 0, lenY)
          pixTot = frame[yNew][xNew]
          if (pixTot < pixTotMax):
            pixTotMax = pixTot
            maxTheta = theta
            xTot = xNew
            yTot = yNew
    
    pixTotList.append(pixTotMax)
    
    if False:
      w = 4
      ym = yTot - w
      yM = yTot + w
      xm = xTot - w
      xM = xTot + w
      if ym < 0:
        ym = 0
      if xm < 0:
        xm = 0
      if yM > lenY + 1:
        yM = lenY + 1
      if xM > lenX + 1:
        xM = lenX + 1
      pixSur = np.min(frame[ym:yM, xm:xM])
    else:
      pixSur = frame[yTot, xTot]

    # Calculates distance between new and old point
    distSubsquentPoints = math.sqrt((xTot - x)**2 + (yTot - y)**2)
    
    if depth + distSubsquentPoints < maxDepth and ((pixSur < pixSurMax) or (depth < hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth)):
      points = appendPoint(xTot, yTot, points)
    else:
      vectX = xTot - x
      vectY = yTot - y
      xTot  = int(x + (maxDepth / (depth + distSubsquentPoints)) * vectX)
      yTot  = int(y + (maxDepth / (depth + distSubsquentPoints)) * vectY)
      points = appendPoint(xTot, yTot, points)
    
    if debug:
      frameDisplay = frame.copy()
      cv2.circle(frameDisplay, (xTot, yTot), 1, (0,0,0),   -1)
      util.showFrame(frameDisplay, title="HeadEmbeddedTailTracking")
    
    newTheta = calculateAngle(x,y,xTot,yTot)
    
    angle = newTheta
    if depth == 0:
      lastFirstTheta = angle
    depth = depth + distSubsquentPoints
    x = xTot
    y = yTot
  
  lenPoints = len(points[0]) - 1
  if points[0, lenPoints-1] == points[0, lenPoints] and points[1, lenPoints-1] == points[1, lenPoints]:
    points = points[:, :len(points[0])-1]
  
  if debug:
    print(np.median(pixTotList), np.mean(pixTotList), pixTotList)
  
  medianPixTotList = np.median(pixTotList)
  
  if hyperparameters["maximumMedianValueOfAllPointsAlongTheTail"] and medianPixTotList > hyperparameters["maximumMedianValueOfAllPointsAlongTheTail"] and not(secondTry):
    # print("second try")
    (points2, lastFirstTheta2, medianPixTotList2) = __insideTrackTail(0, headPosition, frame, np.zeros((2, 0)), (lastFirstTheta + math.pi) % (2 * math.pi), maxDepth, steps, nbList, hyperparameters, debug, lenX, lenY, True)
    if medianPixTotList2 < medianPixTotList:
      return (points2, lastFirstTheta2, medianPixTotList2)
  
  return (points, lastFirstTheta, medianPixTotList)


def trackTail(self, frameROI, headPosition, hyperparameters, wellNumber, frameNumber, lastFirstTheta):
  
  steps   = hyperparameters["steps"]
  nbList  = 10 if hyperparameters["nbList"] == -1 else hyperparameters["nbList"]
  maxDepth = hyperparameters["maxDepth"]

  angle = 0

  points = np.zeros((2, 0))
  
  debug = hyperparameters["debugHeadEmbededFindNextPoints"]
  if debug:
    print("Frame number:", frameNumber + self._firstFrame)
  
  lenX = len(frameROI[0]) - 1
  lenY = len(frameROI) - 1
  
  (points, lastFirstTheta, medianPixTotList) = __insideTrackTail(0, headPosition, frameROI, points, lastFirstTheta, maxDepth, steps, nbList,  hyperparameters, debug, lenX, lenY)
  
  points = np.insert(points, 0, headPosition, axis=1)
  
  if False:
    points = np.transpose(points)
    num_samples = hyperparameters["nbTailPoints"]
    x = points[:, 0]
    y = points[:, 1]
    path_length = np.sum(np.sqrt(np.diff(x)**2 + np.diff(y)**2))
    desired_distance = path_length / (num_samples - 1)
    cumulative_distances = np.cumsum(np.sqrt(np.diff(x)**2 + np.diff(y)**2))
    cumulative_distances = np.insert(cumulative_distances, 0, 0)
    interp_func = interp1d(cumulative_distances, points, axis=0)
    new_distances = np.linspace(0, path_length, num=num_samples)
    resampled_points = interp_func(new_distances)
    points = np.transpose(resampled_points)
  
  output = np.zeros((1, hyperparameters["nbTailPoints"], 2))
  for idx, x in enumerate(points[0]):
    if idx < hyperparameters["nbTailPoints"]:
      output[0][idx][0] = x
      output[0][idx][1] = points[1][idx]
  
  if len(points) < hyperparameters["nbTailPoints"]:
    for idx in range(len(points[0]), hyperparameters["nbTailPoints"]):
      output[0][idx][0] = output[0][len(points[0])-1][0]
      output[0][idx][1] = output[0][len(points[0])-1][1]
  else:
    if len(points) > hyperparameters["nbTailPoints"]:
      output[0][hyperparameters["nbTailPoints"]-1][0] = points[0][len(points)-1]
      output[0][hyperparameters["nbTailPoints"]-1][1] = points[1][len(points)-1]
  
  return output, lastFirstTheta
